# -*- coding: utf-8 -*-

import os.path as op
import numpy as np
import tempfile
import unittest
import shutil

from nipy.modalities.fmri.experimental_paradigm import (EventRelatedParadigm,
                                                        BlockParadigm)

import pyhrf
import pyhrf.paradigm as pdgm


class ParadigmTest(unittest.TestCase):

    def setUp(self):
        self.onsets_1sess = {'c1': [np.array([1.2, 5.6, 7.4])],
                             'c2': [np.array([2.3, 3.6, 6.4, 9.8])]}
        self.durations_1sess = {'c1': [np.array([3., 3., 3.])],
                                'c2': [np.array([4., 4., 4., 4.])], }

        self.onsets_2sess = {'c1': [np.array([1.2, 5.6, 7.4]),
                                    np.array([1.2, 5.6, 7.4, 11.2])],
                             'c2': [np.array([2.3, 3.6, 6.4, 9.8]),
                                    np.array([2.3, 3.6, 6.4, 9.8, 20.7])]}
        self.durations_2sess = {'c1': [np.array([3., 3., 3.]),
                                       np.array([3., 3., 3., 3.])],
                                'c2': [np.array([4., 4., 4., 4.]),
                                       np.array([4., 4., 4., 4., 4.])], }

        tmpDir = tempfile.mkdtemp(prefix='pyhrf_tests',
                                  dir=pyhrf.cfg['global']['tmp_path'])
        self.tmp_dir = tmpDir

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def test_to_nipy_ER(self):
        """
        Test event-related paradigm
        """

        p = pdgm.Paradigm(self.onsets_1sess)
        p_nipy = p.to_nipy_paradigm()
        assert isinstance(p_nipy, EventRelatedParadigm)

    def test_to_nipy_ER_2sess(self):
        """
        Test event-related paradigm
        """

        p = pdgm.Paradigm(self.onsets_2sess)
        p_nipy = p.to_nipy_paradigm()
        assert isinstance(p_nipy, EventRelatedParadigm)

    def test_to_nipy_Block(self):
        """
        Test event-related paradigm
        """

        p = pdgm.Paradigm(self.onsets_1sess, stimDurations=self.durations_1sess)
        p_nipy = p.to_nipy_paradigm()
        assert isinstance(p_nipy, BlockParadigm)

    def test_to_nipy_Block_2sess(self):
        """
        Test event-related paradigm
        """

        p = pdgm.Paradigm(self.onsets_2sess, stimDurations=self.durations_2sess)
        p_nipy = p.to_nipy_paradigm()
        assert isinstance(p_nipy, BlockParadigm)

    def test_to_spm_mat_1st_level(self):
        p = pdgm.Paradigm(self.onsets_2sess, stimDurations=self.durations_2sess)
        p.save_spm_mat_for_1st_level_glm(op.join(self.tmp_dir,
                                                 'conditions.mat'))

    def test_merge_onsets(self):

        o, d = pdgm.merge_onsets(pdgm.onsets_loc, 'audio',
                                 durations=pdgm.durations_loc)

        self.assertIsInstance(o['audio'], list)
        self.assertIsInstance(d['audio'], list)
        self.assertEqual(len(o['audio']), 1)
        self.assertIsInstance(o['audio'][0], np.ndarray)
        self.assertEqual(o['audio'][0].ndim, 1)
        self.assertEqual(o['audio'][0].size, 30)

    def test_onsets_loc_av(self):

        o = pdgm.onsets_loc_av

        self.assertIsInstance(o['audio'], list)
        self.assertEqual(len(o['audio']), 1)
        self.assertIsInstance(o['audio'][0], np.ndarray)
        self.assertEqual(o['audio'][0].ndim, 1)
        self.assertEqual(o['audio'][0].size, 30)



# -*- coding: utf-8 -*-


import numpy as _np
import sys
import copy as CM

debug = 0

def walkCluster(i, links, labels, l, remainingIndexes):
    #print 'walk from ', i
    #cluster.add(i)
    if debug: print '-> treating ...', i
    labels[i] = l
    #visited[i] = True
    if len(links) == 0:
        #print 'no more links'
        return
    #print 'links to :', links[i]
    for e in links[i]:
        if e in remainingIndexes:
            remainingIndexes.remove(e)
            walkCluster(e, links, labels, l, remainingIndexes)


LINK_GRAPH_METHOD = 1
LINK_MAT_METHOD = 0
LINK_CLUST_METHOD = 2

def linkNodes(RefGraph, beta, GraphNodesLabels, GraphLinks, RefGrphNgbhPosi):
    
    for i in xrange(len(RefGraph)):
        for j in xrange(len(RefGraph[i])):
            if debug: print '** %d and %d' %(i,RefGraph[i][j])
            if RefGraph[i][j]>i:  # to avoid two comparisons
                                  # for each edge (non-oriented graph)
                if debug: print '-> treating ...'
                u=(int)(_np.exp(beta)*_np.random.rand()>1. and \
                        GraphNodesLabels[i]==GraphNodesLabels[RefGraph[i][j]])
                if u and debug : print '  link!' 
                GraphLinks[i][j]=u
                
                if RefGrphNgbhPosi==None:
                    for k in xrange(len(RefGraph[RefGraph[i][j]])):
                        if RefGraph[RefGraph[i][j]][k]==i:
                            GraphLinks[RefGraph[i][j]][k]=u
                else:
                    GraphLinks[RefGraph[i][j]][RefGrphNgbhPosi[i][j]]=u



def linkNodesSets(RefGraph, beta, GraphNodesLabels, links, weights=None):
    
    for i in xrange(len(RefGraph)):
        for j in xrange(len(RefGraph[i])):
            if debug: print '** %d and %d' %(i,RefGraph[i][j])
            if RefGraph[i][j]>i:  # to avoid two comparisons for each edge
                                  # (non-oriented graph)
                if debug: print '-> treating ...'
                if weights is None:
                    u = (_np.exp(beta) * (1-_np.random.rand()) >1. and \
                             GraphNodesLabels[i]==GraphNodesLabels[RefGraph[i][j]])
                else:
                    if 0:
                        print 'i:',i,'j:',j
                        print 'RefGraph:', len(RefGraph)
                        print 'RefGraph[i]:', len(RefGraph[i])
                        print 'RefGraph[i][j]:',RefGraph[i][j]
                        print 'weights:', len(weights)
                        print 'weights[i]:', len(weights[i])
                        print 'weights[i][j]:',weights[i][j]
                    u = (_np.exp(beta * weights[i][j]) * (1-_np.random.rand())>1. and \
                             GraphNodesLabels[i]==GraphNodesLabels[RefGraph[i][j]])
                if u:
                    if debug: print '  link!'
                    idxToLink = RefGraph[i][j]
                    links[idxToLink].add(i)
                    links[i].add(idxToLink)
    
def pickLabels(RefGraph, GraphLinks, GraphNodesLabels, NbLabels,
               TempVec, NextTempVec):
    
    for i in xrange(len(GraphNodesLabels)):
        GraphNodesLabels[i]=-1

    for i in xrange(len(RefGraph)):
        if GraphNodesLabels[i]==-1:
            #choice of a label for the connected region
            TempLabel=_np.floor(NbLabels*_np.random.rand())
            #label spread...
            #...init
            TempVec[0]=i
            TempVecSize=1
            GraphNodesLabels[i]=TempLabel

            #...label spread
            while TempVecSize>0:
                NextTempVecSize=0
                for tmp in xrange(TempVecSize):
                    CurntEntry=TempVec[tmp]
                    for CurntNghbdID in xrange(len(RefGraph[CurntEntry])):
                        if (GraphLinks[CurntEntry][CurntNghbdID]==1 and GraphNodesLabels[RefGraph[CurntEntry][CurntNghbdID]]==-1):
                            GraphNodesLabels[RefGraph[CurntEntry][CurntNghbdID]]=TempLabel
                            NextTempVec[NextTempVecSize]=RefGraph[CurntEntry][CurntNghbdID]
                            NextTempVecSize=NextTempVecSize+1

                TempVecSize=NextTempVecSize

                TempVec[0:TempVecSize]=NextTempVec[0:TempVecSize]

def set_cluster_labels(links, labels, nbClasses):
    
    visited = set([])
    for ni, nl in enumerate(links):
        if ni not in visited:
            visited.add(ni)
            l = _np.random.randint(nbClasses)
            labels[ni] = l
            queue = set(nl)
            while len(queue)>0 :
                k = queue.pop()
                if k not in visited:
                    labels[k] = l
                    visited.add(k)
                    queue.update(links[k])


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def SwendsenWangSampler_graph(RefGraph,GraphNodesLabels,beta,NbLabels,
                              GraphLinks=None,RefGrphNgbhPosi=None, method=1,
                              weights=None):
    """
    image sampling with Swendsen-Wang algorithm

    input:

    * RefGraph:
      List which contains the connectivity graph. Each entry represents
      a node of the graph and contains the list of its neighbors entry
      location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is
      the 10th node.
      => There exists i such that RefGraph[10][i]=2
    * GraphNodesLabels:
      list containing the nodes labels. The sampler aims to modify
      its values in function of beta and NbLabels.
    * beta:
      normalization constant
    * NbLabels:
      number of labels (connected voxel pairs are considered if their
      labels are equal)
    * GraphLinks:
      Same shape as RefGraph. Each entry indicates if the link of the
      corresponding edge in RefGraph is considered
      (if yes ...=1 else ...=0).
      This optional list is used as a temporary variable and will be
      modified ! Defining it makes the algorithm faster (no memory
      allocation).
    * RefGrphNgbhPosi:
      Same shape as RefGraph. RefGrphNgbhPosi[i][j] indicates for which
      k is the link to i in RefGraph[RefGraph[i][j]][k]
      This optional list is never modified.

    output:

    * GraphNodesLabels:
      resampled nodes labels. (not returned but modified)
    """
    
    #initializations...
    NodesNb=len(RefGraph)
    
    if GraphLinks==None and method == LINK_MAT_METHOD:
        GraphLinks=CptDefaultGraphLinks(RefGraph)
    
        
        
    #...vectors which will avoid the use of a recursive method for label spreading
    TempVec=_np.zeros(NodesNb,dtype=int)
    NextTempVec=_np.zeros(NodesNb,dtype=int)

    
    #links the voxels
    #print 'RefGraph :'
    #print RefGraph

    if method == LINK_MAT_METHOD:
        if RefGrphNgbhPosi==None:
            RefGrphNgbhPosi=CptRefGrphNgbhPosi(RefGraph)
        linkNodes(RefGraph, beta, GraphNodesLabels, GraphLinks, RefGrphNgbhPosi)
        pickLabels(RefGraph, GraphLinks, GraphNodesLabels, NbLabels,
                   TempVec, NextTempVec)
            
    if method == LINK_GRAPH_METHOD:
        links = range(NodesNb)
        for i in xrange(NodesNb):
            links[i] = set()
        linkNodesSets(RefGraph, beta, GraphNodesLabels, links,
                      weights)
        if debug:
            print 'links :'
            print links
        set_cluster_labels(links, GraphNodesLabels, NbLabels)
        #remainingSites = set(range(NodesNb))
        #while len(remainingSites)>0:
        #    l = _np.random.randint(NbLabels)
        #    i = remainingSites.pop()
        #    walkCluster(i, links, GraphNodesLabels, l, remainingSites)

            
    
#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def Cpt_U_graph(RefGraph,GraphNodesLabels,GraphWeight=None):
    """
    Computes an estimation of U(Graph)

    inputs:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2
    * GraphNodesLabels: list containing the nodes labels.
    * GraphWeight: Same shape as RefGraph. Each entry is the weight of the corresponding
      edge in RefGraph. If not defined the weights are set to 1.0.

    output:

    * U value
    """
    if GraphWeight==None:
        GraphWeight=CptDefaultGraphWeight(RefGraph)
        
    U=0.
    for i in xrange(len(RefGraph)):
        for j in xrange(len(RefGraph[i])):
            if RefGraph[i][j]>i \
                    and GraphNodesLabels[i]==GraphNodesLabels[RefGraph[i][j]] :
                U += GraphWeight[i][j]
    
    return U


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def Cpt_Vec_U_graph(RefGraph,beta,LabelsNb,SamplesNb,
                    GraphWeight=None,GraphNodesLabels=None,
                    GraphLinks=None,RefGrphNgbhPosi=None):
    """
    Computes a given number of U for fields generated according to a given
    normalization constant Beta. Swendsen-Wang sampling is used to generate fields.

    input:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2
    * beta: normalization constant
    * LabelsNb: Labels number
    * SamplesNb: Samples number for the U estimations
    * GraphWeight: Same shape as RefGraph. Each entry is the weight of the corresponding
      edge in RefGraph. If not defined the weights are set to 1.0.
    * GraphNodesLabels: Optional list containing the nodes labels. The sampler aims to modify its values in function of
      beta and NbLabels. At this level this variable is seen as temporary and will be modified. Defining
      it slightly increases the calculation times.
    * GraphLinks: Same shape as RefGraph. Each entry indicates if the link of the corresponding
      edge in RefGraph is considered (if yes ...=1 else ...=0).
      At this level this variable is seen as temporary and will be modified. Defining
      it slightly increases the calculation times.
    * RefGrphNgbhPosi: Same shape as RefGraph. RefGrphNgbhPosi[i][j] indicates for which k is the link to i in
      RefGraph[RefGraph[i][j]][k]. This optional list is never modified.

    output:

    * VecU: Vector of size SamplesNb containing the U computations
    """
    #initialization
    if GraphWeight is None:
        GraphWeight=CptDefaultGraphWeight(RefGraph)
    
    if GraphNodesLabels==None:
        GraphNodesLabels=CptDefaultGraphNodesLabels(RefGraph)
    else:
        for i in xrange(len(GraphNodesLabels)):
            GraphNodesLabels[i]=0
    
    if GraphLinks==None:
        GraphLinks=CptDefaultGraphLinks(RefGraph)
    
    if RefGrphNgbhPosi==None:
        RefGrphNgbhPosi=CptRefGrphNgbhPosi(RefGraph)
        
    #all estimates of ImagLoc will then be significant in the expectation calculation (initial field is homogeneous)
    SwendsenWangSampler_graph(RefGraph,GraphNodesLabels,beta,LabelsNb,
                              GraphLinks=GraphLinks,RefGrphNgbhPosi=RefGrphNgbhPosi,
                              weights=GraphWeight)
    
    #estimation
    VecU=_np.zeros(SamplesNb)
    
    for i in xrange(SamplesNb):
        SwendsenWangSampler_graph(RefGraph,GraphNodesLabels,beta,
                                  LabelsNb,GraphLinks=GraphLinks,RefGrphNgbhPosi=RefGrphNgbhPosi,
                                  weights=GraphWeight)
        
        VecU[i]=Cpt_U_graph(RefGraph,GraphNodesLabels,GraphWeight=GraphWeight)
    
    return VecU



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def CptRefGrphNgbhPosi(RefGraph):
    """
    computes the critical list CptRefGrphNgbhPosi from RefGraph

    imput:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2
      
    output:

    * RefGrphNgbhPosi: Same shape as RefGraph. RefGrphNgbhPosi[i][j] indicates for which k is the link to i in
      RefGraph[RefGraph[i][j]][k]. It makes algorithms which run through the graph much faster
      since it avoids a critical loop.
    """
    
    RefGrphNgbhPosi=[]
    for i in xrange(len(RefGraph)):
        RefGrphNgbhPosi.append([])
        
    for CurntEntry in xrange(len(RefGraph)):
        for CurntNghbdID in xrange(len(RefGraph[CurntEntry])):
            CurntNghbd=RefGraph[CurntEntry][CurntNghbdID]
            for k in xrange(len(RefGraph[CurntNghbd])):
                if RefGraph[CurntNghbd][k]==CurntEntry:
                    RefGrphNgbhPosi[CurntEntry].append(k)
    
    return RefGrphNgbhPosi


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def CptDefaultGraphNodesLabels(RefGraph):
    """
    computes a default list GraphNodesLabels from RefGraph

    input:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2

    output:

    * GraphNodesLabels: List containing the nodes labels (consiedered for the computation of U). The sampler
      aims to modify its values in function of beta and NbLabels.
    """
    
    GraphNodesLabels=[]
    for i in xrange(len(RefGraph)):
        GraphNodesLabels.append(0)
    
    return GraphNodesLabels

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def CptDefaultGraphLinks(RefGraph):
    """
    computes a default list GraphLinks from RefGraph

    input:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2

    output:

    * GraphLinks: Same shape as RefGraph. Each entry indicates whether the link of the corresponding
      edge in RefGraph is considered or not in Swensdsen-Wang Sampling (1 -> yes / 0 -> no).
    """
    
    GraphLinks=[]
    for i in xrange(len(RefGraph)):
        GraphLinks.append([])
        for j in xrange(len(RefGraph[i])):
            GraphLinks[i].append(1)
    
    return GraphLinks

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def CptDefaultGraphWeight(RefGraph):
    """
    computes a default list GraphWeight from RefGraph. Each edge weight is set to 1.0.

    input:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2
      
    output:

    * GraphWeight: Same shape as RefGraph. Each entry is the weight of the corresponding edge in RefGraph
    """
    GraphWeight = range(len(RefGraph))
    for i in xrange(len(RefGraph)):
        GraphWeight[i] = [1.] * len(RefGraph[i])
    return GraphWeight

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def GraphBetaMix(RefGraph,GraphNodesLabels,beta=0.5,NbLabels=2,NbIt=5,weights=None):
    """
    Generate a partition in GraphNodesLabels with respect to beta.

    input:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2
    * GraphNodesLabels: list containing the nodes labels.
    * beta: correlation factor for all conditions
    * NbLabels: number of labels in all site conditions
    * NbIt: number of sampling steps with SwendsenWangSampler_graph

    output:

    * GraphNodesLabels: sampled GraphNodesLabels (not returned but modified)
    """
    
    for i in xrange(NbIt):
        SwendsenWangSampler_graph(RefGraph,GraphNodesLabels,beta,NbLabels,weights)
        

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def ImageToGraph(Image,Mask,LabelOI=1,ConnectivityType=6):
    """
    Computes the connectivity graph of an image under a 3D mask voxels.

    inputs:

    * Image: the 3D image (label field).
    * Mask: corresponding 3D mask.
    * LabelOI: Voxels of Mask containing the label LabelOI are those considered.
    * ConnectivityType: controles the connectivity considered in the graph.
      ConnectivityType=26 : 26-connectivity
      ConnectivityType=18 : 18-connectivity
      ConnectivityType=6 : 6-connectivity

    outputs:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2
    * GraphWeight: Same shape as RefGraph. Each entry is the weight of the corresponding
      edge in RefGraph
    * GraphNodesCoord: Coordinates of each node in Mask
    * GraphNodesLabels: list containing the nodes labels.
    """
    
    #initialization
    NBZ=Image.shape[0]
    NBY=Image.shape[1]
    NBX=Image.shape[2]
    NBZb=Mask.shape[0]
    NBYb=Mask.shape[1]
    NBXb=Mask.shape[2]
    
    if (NBZ!=NBZb) or (NBY!=NBYb) or (NBX!=NBXb):
        print "!!! image and mask shapes do not correspond  !!!"
    
    #label the 1-voxels of MaskID
    MaskID=_np.int16(Mask.copy())
    ID=0
    for i in xrange(NBZ):
        for j in xrange(NBY):
            for k in xrange(NBX):
                if MaskID[i,j,k]!=LabelOI:
                    MaskID[i,j,k]=-1
                else:
                    MaskID[i,j,k]=ID
                    ID=ID+1
    
    #creates the Mask under its graph formulation
    RefGraph=[]
    GraphWeight=[]
    GraphNodesCoord=[]
    GraphNodesLabels=[]
    
    ID=0
    for i in xrange(NBZ):
        for j in xrange(NBY):
            for k in xrange(NBX):
                if MaskID[i,j,k]!=-1:
                    #tests the consistency with the previous loops
                    if MaskID[i,j,k]==ID:
                        ID=ID+1
                    else:
                        print 'consistency problem in the graph creation'
                    
                    #creation of the MaskID[i,j,k]'th entry in RefGraph
                    RefGraph.append([])
                    GraphWeight.append([])
                    GraphNodesCoord.append([k,j,i])
                    GraphNodesLabels.append(Image[i,j,k])
                    
                    if ConnectivityType==6 or ConnectivityType==18 or ConnectivityType==26:
                        if (i<NBZ-1):
                            if Mask[i+1,j,k]==1:
                                RefGraph[-1].append(MaskID[i+1,j,k])
                                GraphWeight[-1].append(1.)
                        if (i>0):
                            if Mask[i-1,j,k]==1:
                                RefGraph[-1].append(MaskID[i-1,j,k])
                                GraphWeight[-1].append(1.)
                        if (j<NBY-1):
                            if Mask[i,j+1,k]==1:
                                RefGraph[-1].append(MaskID[i,j+1,k])
                                GraphWeight[-1].append(1.)
                        if (j>0):
                            if Mask[i,j-1,k]==1:
                                RefGraph[-1].append(MaskID[i,j-1,k])
                                GraphWeight[-1].append(1.)
                        if (k<NBX-1):
                            if Mask[i,j,k+1]==1:
                                RefGraph[-1].append(MaskID[i,j,k+1])
                                GraphWeight[-1].append(1.)
                        if (k>0):
                            if Mask[i,j,k-1]==1:
                                RefGraph[-1].append(MaskID[i,j,k-1])
                                GraphWeight[-1].append(1.)
                        
                    if ConnectivityType==18 or ConnectivityType==26:
                        if (i<NBZ-1) and (j<NBY-1):
                            if Mask[i+1,j+1,k]==1:
                                RefGraph[-1].append(MaskID[i+1,j+1,k])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i<NBZ-1) and (j>0):
                            if Mask[i+1,j-1,k]==1:
                                RefGraph[-1].append(MaskID[i+1,j-1,k])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i<NBZ-1) and (k<NBX-1):
                            if Mask[i+1,j,k+1]==1:
                                RefGraph[-1].append(MaskID[i+1,j,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i<NBZ-1) and (k>0):
                            if Mask[i+1,j,k-1]==1:
                                RefGraph[-1].append(MaskID[i+1,j,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i>0) and (j<NBY-1):
                            if Mask[i-1,j+1,k]==1:
                                RefGraph[-1].append(MaskID[i-1,j+1,k])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i>0) and (j>0):
                            if Mask[i-1,j-1,k]==1:
                                RefGraph[-1].append(MaskID[i-1,j-1,k])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i>0) and (k<NBX-1):
                            if Mask[i-1,j,k+1]==1:
                                RefGraph[-1].append(MaskID[i-1,j,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i>0) and (k>0):
                            if Mask[i-1,j,k-1]==1:
                                RefGraph[-1].append(MaskID[i-1,j,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (j<NBY-1) and (k<NBX-1):
                            if Mask[i,j+1,k+1]==1:
                                RefGraph[-1].append(MaskID[i,j+1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (j<NBY-1) and (k>0):
                            if Mask[i,j+1,k-1]==1:
                                RefGraph[-1].append(MaskID[i,j+1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (j>0) and (k<NBX-1):
                            if Mask[i,j-1,k+1]==1:
                                RefGraph[-1].append(MaskID[i,j-1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (j>0) and (k>0):
                            if Mask[i,j-1,k-1]==1:
                                RefGraph[-1].append(MaskID[i,j-1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                    if ConnectivityType==26:
                        if (i<NBZ-1) and (j<NBY-1) and (k<NBX-1):
                            if Mask[i+1,j+1,k+1]==1:
                                RefGraph[-1].append(MaskID[i+1,j+1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i<NBZ-1) and (j<NBY-1) and (k>0):
                            if Mask[i+1,j+1,k-1]==1:
                                RefGraph[-1].append(MaskID[i+1,j+1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i<NBZ-1) and (j>0) and (k<NBX-1):
                            if Mask[i+1,j-1,k+1]==1:
                                RefGraph[-1].append(MaskID[i+1,j-1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i<NBZ-1) and (j>0) and (k>0):
                            if Mask[i+1,j-1,k-1]==1:
                                RefGraph[-1].append(MaskID[i+1,j-1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i>0) and (j<NBY-1) and (k<NBX-1):
                            if Mask[i-1,j+1,k+1]==1:
                                RefGraph[-1].append(MaskID[i-1,j+1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i>0) and (j<NBY-1) and (k>0):
                            if Mask[i-1,j+1,k-1]==1:
                                RefGraph[-1].append(MaskID[i-1,j+1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i>0) and (j>0) and (k<NBX-1):
                            if Mask[i-1,j-1,k+1]==1:
                                RefGraph[-1].append(MaskID[i-1,j-1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i>0) and (j>0) and (k>0):
                            if Mask[i-1,j-1,k-1]==1:
                                RefGraph[-1].append(MaskID[i-1,j-1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
    
    return RefGraph,GraphWeight,GraphNodesCoord,GraphNodesLabels


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def MaskToGraph(Mask,LabelOI=1,ConnectivityType=6):
    """
    Computes the connectivity graph of in 3D mask voxels.

    inputs:

    * Mask: 3D mask.
    * LabelOI: Voxels of Mask containing the label LabelOI are those considered.
    * ConnectivityType: controles the connectivity considered in the graph.
      ConnectivityType=26 : 26-connectivity
      ConnectivityType=18 : 18-connectivity
      ConnectivityType=6 : 6-connectivity

    outputs:

    * RefGraph: List which contains the connectivity graph. Each entry represents a node of the graph
      and contains the list of its neighbors entry location in the graph.
      ex: RefGraph[2][3]=10 means 3rd neighbour of the 2nd node is the 10th node.
      => There exists i such that RefGraph[10][i]=2
    * GraphWeight: Same shape as RefGraph. Each entry is the weight of the corresponding
      edge in RefGraph
    * GraphNodesCoord: Coordinates of each node in Mask
    """
    
    #initialization
    NBZ=Mask.shape[0]
    NBY=Mask.shape[1]
    NBX=Mask.shape[2]
    
    
    #label the 1-voxels of MaskID
    MaskID=_np.int16(Mask.copy())
    ID=0
    for i in xrange(NBZ):
        for j in xrange(NBY):
            for k in xrange(NBX):
                if MaskID[i,j,k]!=LabelOI:
                    MaskID[i,j,k]=-1
                else:
                    MaskID[i,j,k]=ID
                    ID=ID+1
    
    #creates the Mask under its graph formulation
    RefGraph=[]
    GraphWeight=[]
    GraphNodesCoord=[]
    
    ID=0
    for i in xrange(NBZ):
        for j in xrange(NBY):
            for k in xrange(NBX):
                if MaskID[i,j,k]!=-1:
                    #tests the consistency with the previous loops
                    if MaskID[i,j,k]==ID:
                        ID=ID+1
                    else:
                        print 'consistency problem in the graph creation'
                    
                    #creation of the MaskID[i,j,k]'th entry in RefGraph
                    RefGraph.append([])
                    GraphWeight.append([])
                    GraphNodesCoord.append([k,j,i])
                    
                    if ConnectivityType==6 or ConnectivityType==18 or ConnectivityType==26:
                        if (i<NBZ-1):
                            if Mask[i+1,j,k]==1:
                                RefGraph[-1].append(MaskID[i+1,j,k])
                                GraphWeight[-1].append(1.)
                        if (i>0):
                            if Mask[i-1,j,k]==1:
                                RefGraph[-1].append(MaskID[i-1,j,k])
                                GraphWeight[-1].append(1.)
                        if (j<NBY-1):
                            if Mask[i,j+1,k]==1:
                                RefGraph[-1].append(MaskID[i,j+1,k])
                                GraphWeight[-1].append(1.)
                        if (j>0):
                            if Mask[i,j-1,k]==1:
                                RefGraph[-1].append(MaskID[i,j-1,k])
                                GraphWeight[-1].append(1.)
                        if (k<NBX-1):
                            if Mask[i,j,k+1]==1:
                                RefGraph[-1].append(MaskID[i,j,k+1])
                                GraphWeight[-1].append(1.)
                        if (k>0):
                            if Mask[i,j,k-1]==1:
                                RefGraph[-1].append(MaskID[i,j,k-1])
                                GraphWeight[-1].append(1.)
                        
                    if ConnectivityType==18 or ConnectivityType==26:
                        if (i<NBZ-1) and (j<NBY-1):
                            if Mask[i+1,j+1,k]==1:
                                RefGraph[-1].append(MaskID[i+1,j+1,k])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i<NBZ-1) and (j>0):
                            if Mask[i+1,j-1,k]==1:
                                RefGraph[-1].append(MaskID[i+1,j-1,k])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i<NBZ-1) and (k<NBX-1):
                            if Mask[i+1,j,k+1]==1:
                                RefGraph[-1].append(MaskID[i+1,j,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i<NBZ-1) and (k>0):
                            if Mask[i+1,j,k-1]==1:
                                RefGraph[-1].append(MaskID[i+1,j,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i>0) and (j<NBY-1):
                            if Mask[i-1,j+1,k]==1:
                                RefGraph[-1].append(MaskID[i-1,j+1,k])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i>0) and (j>0):
                            if Mask[i-1,j-1,k]==1:
                                RefGraph[-1].append(MaskID[i-1,j-1,k])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i>0) and (k<NBX-1):
                            if Mask[i-1,j,k+1]==1:
                                RefGraph[-1].append(MaskID[i-1,j,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (i>0) and (k>0):
                            if Mask[i-1,j,k-1]==1:
                                RefGraph[-1].append(MaskID[i-1,j,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (j<NBY-1) and (k<NBX-1):
                            if Mask[i,j+1,k+1]==1:
                                RefGraph[-1].append(MaskID[i,j+1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (j<NBY-1) and (k>0):
                            if Mask[i,j+1,k-1]==1:
                                RefGraph[-1].append(MaskID[i,j+1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (j>0) and (k<NBX-1):
                            if Mask[i,j-1,k+1]==1:
                                RefGraph[-1].append(MaskID[i,j-1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                        if (j>0) and (k>0):
                            if Mask[i,j-1,k-1]==1:
                                RefGraph[-1].append(MaskID[i,j-1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(2.))
                    if ConnectivityType==26:
                        if (i<NBZ-1) and (j<NBY-1) and (k<NBX-1):
                            if Mask[i+1,j+1,k+1]==1:
                                RefGraph[-1].append(MaskID[i+1,j+1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i<NBZ-1) and (j<NBY-1) and (k>0):
                            if Mask[i+1,j+1,k-1]==1:
                                RefGraph[-1].append(MaskID[i+1,j+1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i<NBZ-1) and (j>0) and (k<NBX-1):
                            if Mask[i+1,j-1,k+1]==1:
                                RefGraph[-1].append(MaskID[i+1,j-1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i<NBZ-1) and (j>0) and (k>0):
                            if Mask[i+1,j-1,k-1]==1:
                                RefGraph[-1].append(MaskID[i+1,j-1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i>0) and (j<NBY-1) and (k<NBX-1):
                            if Mask[i-1,j+1,k+1]==1:
                                RefGraph[-1].append(MaskID[i-1,j+1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i>0) and (j<NBY-1) and (k>0):
                            if Mask[i-1,j+1,k-1]==1:
                                RefGraph[-1].append(MaskID[i-1,j+1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i>0) and (j>0) and (k<NBX-1):
                            if Mask[i-1,j-1,k+1]==1:
                                RefGraph[-1].append(MaskID[i-1,j-1,k+1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
                        if (i>0) and (j>0) and (k>0):
                            if Mask[i-1,j-1,k-1]==1:
                                RefGraph[-1].append(MaskID[i-1,j-1,k-1])
                                GraphWeight[-1].append(1./_np.sqrt(3.))
    
    return RefGraph,GraphWeight,GraphNodesCoord



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def GraphToImage(GraphNodesCoord,GraphNodesLabels,NBZ,NBY,NBX):
    """
    Computes a 3D image from a connectivity graph.

    input:

    * GraphNodesCoord: Coordinates of each node in Mask
    * GraphNodesLabels: Nodes labels. For example, GraphNodesLabels[i] is the label of node i.
    * NBZ: image size on Z axis
    * NBY: image size on Y axis
    * NBX: image size on X axis

    output:

    * Image
    """
    
    Image=_np.zeros((NBZ,NBY,NBX),dtype=int)
    
    for i in xrange(len(GraphNodesCoord)):
        Image[GraphNodesCoord[i][2],GraphNodesCoord[i][1],GraphNodesCoord[i][0]]=GraphNodesLabels[i]
        
    return Image



# -*- coding: utf-8 -*-

import logging

from time import time
from collections import OrderedDict

import numpy as np
import nibabel

from pyhrf.ndarray import xndarray
from pyhrf.ui.jde import JDEAnalyser
from pyhrf.vbjde.vem_tools import roc_curve
from pyhrf.vbjde.vem_bold import jde_vem_bold
from pyhrf.vbjde.vem_bold_constrained import Main_vbjde_Python_constrained
from pyhrf.xmlio import XmlInitable
from pyhrf.tools import format_duration


logger = logging.getLogger(__name__)


def change_dim(labels):
    """Change labels dimension from ``(ncond, nclass, nvox)`` to ``(nclass, ncond, nvox)``."""
    ncond = labels.shape[0]
    nclass = labels.shape[1]
    nvox = labels.shape[2]
    newlabels = np.zeros((nclass, ncond, nvox))
    for cond in xrange(ncond):
        for clas in xrange(nclass):
            for vox in xrange(nvox):
                newlabels[clas][cond][vox] = labels[cond][clas][vox]

    return newlabels


class JDEVEMAnalyser(JDEAnalyser):
    """Class that handles parcel-wise which is done according to the input data parcellation by default, and also takes
    care of merging parcel-specific outputs at the end of the JDE VEM analysis.
        
    Args:
        hrfDuration (float): duration of the HRF in seconds.
        sigmaH (float): initial HRF variance.
        fast (bool): running fast VEM with C extensions. 
        computeContrast (bool): compute the contrasts defined in contrasts input.
        nbClasses (int): number of classes for the response levels. 
        PLOT (bool): plotting flag for convergence curves.
        nItMax (int): maximum number of iterations.
        nItMin (int): minimum number of iterations.
        scale (bool): divide the data fidelity term during m_h step by the number of voxels. 
        beta (float): initial value of spatial Potts regularization parameter. 
        estimateSigmaH (bool): estimate or not the HRF variance. 
        estimateHRF (bool): estimate or not the HRF. 
        TrueHrfFlag (bool): if true, HRF will be fixed to the simulated value. 
        HrfFilename (str): filename of the output HRF. 
        estimateDrifts (bool): if true drifts are estimated, otherwise drifts are marginalized. 
        hyper_prior_sigma_H (int): hyper-prior on sigma_H. If zero, no prior is applied.
        dt (float): time resolution of the estimated HRF (in seconds). 
        estimateBeta (bool): estimate or not the Potts spatial regularization parameter. 
        contrasts (:obj:`~collections.OrderedDict`): contrasts to be evaluated. 
        simulation (bool): indicates whether the run corresponds to a simulation example or not. 
        estimateLabels (bool): estimate or not the labels.
        LabelsFilename (str): filename containing the labels.
        MFapprox (bool): using the Mean Field approximation in labels estimation.  
        estimateMixtParam (bool): estimate or not the mixture parameters. 
        constrained (bool): if true, the following constraints are added: positivity and norm = 1.
        InitVar (float): initial value of active and inactive gaussian variances.
        InitMean (float): initial value of active gaussian means. 
        MiniVemFlag (bool): estimate or not the best initialization of MixtParam and gamma_h. 
        NbItMiniVem (int): number of iterations in Mini VEM algorithm. 
        zero_constraint (bool): if true, add zeros at the beginning and the end of the HRF.
        output_drifts (bool): save the estimated drifts.
        drifts_type (str): type of the drift basis ('poly' or 'cos'). 
    """

    parametersComments = {
        'dt': 'time resolution of the estimated HRF (in seconds)',
        'hrfDuration': 'duration of the HRF (in seconds)',
        'sigmaH': 'initial HRF variance',
        'fast': 'running fast VEM with C extensions',
        'nbClasses': 'number of classes for the response levels',
        'PLOT': 'plotting flag for convergence curves',
        'nItMax': 'maximum number of iterations',
        'nItMin': 'minimum number of iterations',
        'scale': 'if true, the data fidelity term is divide by the number of voxels, otherwise it does nothing',
        'beta': 'initial value of spatial Potts regularization parameter',
        'simulation': 'indicates whether the run corresponds to a simulation example or not',
        'estimateSigmaH': 'estimate or not the HRF variance',
        'estimateHRF': 'estimate or not the HRF',
        'TrueHrfFlag': 'if true, HRF will be fixed to the simulated value',
        'HrfFilename': 'HRF filename',
        'estimateBeta': 'estimate or not the Potts spatial regularization parameter',
        'estimateDrifts': 'explicit drift estimation (if false, drifts are marginalized',
        'drifts_type': 'type of the drift basis (default="polynomial")',
        'contrasts': 'contrasts to be evaluated',
        'hyper_prior_sigma_H': 'parameter of the hyper-prior on sigma_H (if zero, no prior is applied)',
        'estimateLabels': 'estimate or not the labels',
        'LabelsFilename': 'labels filename',
        'MFapprox': 'using of the Mean Field approximation in labels estimation',
        'estimateMixtParam': 'estimate or not the mixture parameters',
        'InitVar': 'initial value of active and inactive gaussian variances',
        'InitMean': 'initial value of active gaussian means',
        'MiniVemFlag': 'if true, estimate the best initialisation of MixtParam and gamma_h',
        'NbItMiniVem': 'number of iterations in Mini VEM algorithm',
        'constrained': 'adding constrains: positivity and norm = 1 ',
        }

    parametersToShow = ['dt', 'hrfDuration', 'nItMax', 'nItMin',
                        'estimateSigmaH', 'estimateHRF', 'TrueHrfFlag',
                        'HrfFilename', 'estimateBeta', 'estimateLabels',
                        'LabelsFilename', 'MFapprox', 'estimateDrifts',
                        'estimateMixtParam', 'InitVar', 'InitMean', 'scale',
                        'nbClasses', 'fast', 'PLOT', 'sigmaH', 'contrasts',
                        'hyper_prior_sigma_H', 'constrained', 'simulation',
                        'MiniVemFlag', 'NbItMiniVem']

    def __init__(self, hrfDuration=25., sigmaH=0.1, fast=True,
                 computeContrast=True, nbClasses=2, PLOT=False, nItMax=100,
                 nItMin=1, scale=False, beta=1.0, estimateSigmaH=True,
                 estimateHRF=True, TrueHrfFlag=False, HrfFilename='hrf.nii',
                 estimateDrifts=True, hyper_prior_sigma_H=1000, dt=.6,
                 estimateBeta=True, contrasts=None, simulation=False,
                 estimateLabels=True, LabelsFilename=None,
                 MFapprox=False, estimateMixtParam=True, constrained=False,
                 InitVar=0.5, InitMean=2.0, MiniVemFlag=False, NbItMiniVem=5,
                 zero_constraint=True, output_drifts=False, drifts_type="poly"):

        XmlInitable.__init__(self)
        JDEAnalyser.__init__(self, outputPrefix='jde_vem_')

        # Important thing : all parameters must have default values
        self.dt = dt
        self.hrfDuration = hrfDuration
        self.nbClasses = nbClasses
        self.nItMax = nItMax
        self.estimateSigmaH = estimateSigmaH
        self.scale = scale
        self.estimateDrifts = estimateDrifts
        self.PLOT = PLOT
        self.fast = fast
        self.simulation = simulation
        self.beta = beta
        self.sigmaH = sigmaH
        self.estimateHRF = estimateHRF
        self.TrueHrfFlag = TrueHrfFlag
        self.HrfFilename = HrfFilename
        self.nItMin = nItMin
        self.estimateBeta = estimateBeta
        self.estimateMixtParam = estimateMixtParam
        self.estimateLabels = estimateLabels
        self.LabelsFilename = LabelsFilename
        self.MFapprox = MFapprox
        self.InitVar = InitVar
        self.InitMean = InitMean
        self.MiniVemFlag = MiniVemFlag
        self.NbItMiniVem = NbItMiniVem
        self.contrasts = OrderedDict() if contrasts is None else contrasts
        self.computeContrast = computeContrast
        self.hyper_prior_sigma_H = hyper_prior_sigma_H
        self.constrained = constrained
        self.zero_constraint = zero_constraint
        self.output_drifts = output_drifts
        self.drifts_type = drifts_type

        logger.info("VEM analyzer:")
        logger.info(" - estimate sigma H: %s", str(self.estimateSigmaH))
        logger.info(" - init sigma H: %f", self.sigmaH)
        logger.info(" - hyper_prior_sigma_H: %f", self.hyper_prior_sigma_H)
        logger.info(" - estimate drift: %s", str(self.estimateDrifts))

    def analyse_roi(self, roiData):
        """ROI analysis of the fMRI data.
        
        Args:
            roiData (:obj:`~pyhrf.core.FmriData`): fMRI data to be analyzed.  

        Returns:
            :obj:`dict` of :obj:`~pyhrf.ndarray.xndarray`: packed outputs.
            
        """
        data = roiData.bold
        Onsets = roiData.get_joined_onsets()
        durations = roiData.get_joined_durations()
        TR = roiData.tr
        # K = 2 #number of classes
        nvox = roiData.get_nb_vox_in_mask()
        scale = nvox if self.scale else 1
        rid = roiData.get_roi_id()
        logger.info("JDE VEM - roi %d, nvox=%d, nconds=%d, nItMax=%d", rid, nvox, len(Onsets), self.nItMax)

        self.contrasts.pop('dummy_example', None)
        cNames = roiData.paradigm.get_stimulus_names()
        graph = roiData.get_graph()

        t_start = time()

        if self.fast:
            logger.info("fast VEM with drift estimation" + ("and a constraint"*self.constrained))

            (nb_iter, nrls_mean, hrf_mean, hrf_covar, labels_proba, noise_var, nrls_class_mean, nrls_class_var, beta,
             drift_coeffs, drift, contrasts_mean, contrasts_var, _, _, nrls_covar, _, density_ratio,
             density_ratio_cano, density_ratio_diff, density_ratio_prod, ppm_a_nrl, ppm_g_nrl, ppm_a_contrasts,
             ppm_g_contrasts, variation_coeff, free_energy, free_energy_crit, beta_list, delay_of_response,
             delay_of_undershoot, dispersion_of_response, dispersion_of_undershoot, ratio_resp_under,
             delay) = jde_vem_bold(graph, data, Onsets, durations, self.hrfDuration, self.nbClasses, TR, self.beta,
                                   self.dt, self.estimateSigmaH, self.sigmaH, self.nItMax, self.nItMin,
                                   self.estimateBeta, self.contrasts, self.computeContrast, self.hyper_prior_sigma_H,
                                   self.estimateHRF, constrained=self.constrained, zero_constraint=self.zero_constraint,
                                   drifts_type=self.drifts_type)

        elif self.estimateDrifts:  # if not self.fast
            logger.info("not fast VEM")
            logger.info("NOT WORKING")
            (nrls_mean, hrf_mean, labels_proba, noise_var, nrls_class_mean, nrls_class_var, beta, drift_coeffs,
             drift) = Main_vbjde_Python_constrained(graph, data, Onsets, self.hrfDuration, self.nbClasses, TR,
                                                    self.beta, self.dt, scale, self.estimateSigmaH, self.sigmaH,
                                                    self.nItMax, self.nItMin, self.estimateBeta, self.PLOT)

        # Plot analysis duration
        self.analysis_duration = time() - t_start
        logger.info('JDE VEM analysis took: %s', format_duration(self.analysis_duration))

        if self.fast:
            # OUTPUTS: Pack all outputs within a dict
            outputs = {}
            hrf_time = np.arange(len(hrf_mean)) * self.dt

            # axes_names = ['iteration']
            # axes_domains = {'iteration':np.arange(FreeEnergy.shape[0])}
            # outputs['FreeEnergy'] = xndarray(FreeEnergy,
            #                             axes_names=axes_names,
            #                             axes_domains=axes_domains)

            outputs['hrf'] = xndarray(hrf_mean, axes_names=['time'],
                                      axes_domains={'time': hrf_time},
                                      value_label="HRF")

            domCondition = {'condition': cNames}
            outputs['nrls'] = xndarray(nrls_mean.transpose(), value_label="nrls",
                                       axes_names=['condition', 'voxel'],
                                       axes_domains=domCondition)

            ad = {'condition': cNames, 'condition2': Onsets.keys()}

            outputs['Sigma_nrls'] = xndarray(nrls_covar, value_label="Sigma_NRLs",
                                             axes_names=['condition', 'condition2', 'voxel'],
                                             axes_domains=ad)

            outputs['nb_iter'] = xndarray(np.array([nb_iter]), value_label="nb_iter")

            outputs['beta'] = xndarray(beta, value_label="beta",
                                       axes_names=['condition'],
                                       axes_domains=domCondition)

            nbc, nbv = len(cNames), nrls_mean.shape[0]
            repeatedBeta = np.repeat(beta, nbv).reshape(nbc, nbv)
            outputs['beta_mapped'] = xndarray(repeatedBeta, value_label="beta",
                                              axes_names=['condition', 'voxel'],
                                              axes_domains=domCondition)

            repeated_hrf = np.repeat(hrf_mean, nbv).reshape(-1, nbv)
            outputs["hrf_mapped"] = xndarray(repeated_hrf, value_label="HRFs",
                                             axes_names=["time", "voxel"],
                                             axes_domains={"time": hrf_time})

            repeated_hrf_covar = np.repeat(np.diag(hrf_covar), nbv).reshape(-1, nbv)
            outputs["hrf_variance_mapped"] = xndarray(repeated_hrf_covar,
                                                      value_label="HRFs covariance",
                                                      axes_names=["time", "voxel"],
                                                      axes_domains={"time": hrf_time})

            outputs['roi_mask'] = xndarray(np.zeros(nbv)+roiData.get_roi_id(),
                                           value_label="ROI",
                                           axes_names=['voxel'])

            outputs["density_ratio"] = xndarray(np.zeros(nbv)+density_ratio,
                                                value_label="Density Ratio to zero",
                                                axes_names=["voxel"])

            outputs["density_ratio_cano"] = xndarray(np.zeros(nbv)+density_ratio_cano,
                                                     value_label="Density Ratio to canonical",
                                                     axes_names=["voxel"])

            outputs["density_ratio_diff"] = xndarray(np.zeros(nbv)+density_ratio_diff,
                                                     value_label="Density Ratio to canonical",
                                                     axes_names=["voxel"])

            outputs["density_ratio_prod"] = xndarray(np.zeros(nbv)+density_ratio_prod,
                                                     value_label="Density Ratio to canonical",
                                                     axes_names=["voxel"])

            outputs["variation_coeff"] = xndarray(np.zeros(nbv)+variation_coeff,
                                                  value_label="Coefficient of variation of the HRF",
                                                  axes_names=["voxel"])

            free_energy = np.concatenate((np.asarray(free_energy), np.zeros((self.nItMax - len(free_energy)))))
            free_energy[free_energy == 0.] = np.nan
            free_energy = np.repeat(free_energy, nbv).reshape(-1, nbv)
            outputs["free_energy"] = xndarray(free_energy,
                                              value_label="free energy",
                                              axes_names=["time", "voxel"])

            if self.estimateHRF:  # if self.fast
                fitting_parameters = {
                    "hrf_fit_delay_of_response":  delay_of_response,
                    "hrf_fit_delay_of_undershoot":  delay_of_undershoot,
                    "hrf_fit_dispersion_of_response":  dispersion_of_response,
                    "hrf_fit_dispersion_of_undershoot":  dispersion_of_undershoot,
                    "hrf_fit_ratio_response_undershoot": ratio_resp_under,
                    "hrf_fit_delay": delay,
                }
                affine = np.eye(4)
                for param_name in fitting_parameters:
                    header = nibabel.Nifti1Header()
                    description = param_name[8:].replace("_", " ").capitalize()
                    outputs[param_name] = xndarray(
                        np.zeros(nbv)+fitting_parameters[param_name],
                        value_label=description + " of the fitted estimated HRF",
                        axes_names=["voxel"], meta_data=(affine, header)
                    )
                    outputs[param_name].meta_data[1]["descrip"] = description

            nrls_mean = nrls_mean.transpose()

            nvox = nrls_mean.shape[1]

            # nbconds = nrls_mean.shape[0]
            # h = hrf_mean
            # ah = np.zeros((h.shape[0], nvox, nbconds))

            mixtp = np.zeros((roiData.nbConditions, self.nbClasses, 2))
            mixtp[:, :, 0] = nrls_class_mean
            mixtp[:, :, 1] = np.sqrt(nrls_class_var)

            an = ['condition', 'Act_class', 'component']
            ad = {'Act_class': ['inactiv', 'activ'],
                  'condition': cNames,
                  'component': ['mean', 'var']}
            outputs['mixt_p'] = xndarray(mixtp, axes_names=an, axes_domains=ad)

            ad = {'class': ['inactiv', 'activ'],
                  'condition': cNames}
            outputs['labels'] = xndarray(labels_proba, value_label="Labels",
                                         axes_names=['condition', 'class', 'voxel'],
                                         axes_domains=ad)
            outputs['noise_var'] = xndarray(noise_var, value_label="noise_var",
                                            axes_names=['voxel'])

            if self.estimateDrifts and self.output_drifts:
                outputs['drift_coeff'] = xndarray(drift_coeffs, value_label="Drift",
                                                  axes_names=['coeff', 'voxel'])
                outputs['drift'] = xndarray(drift, value_label="Delta BOLD",
                                            axes_names=['time', 'voxel'])

            affine = np.eye(4)
            for condition_nb, condition_name in enumerate(cNames):
                header = nibabel.Nifti1Header()
                outputs["ppm_a_nrl_"+condition_name] = xndarray(ppm_a_nrl[:, condition_nb],
                                                                value_label="PPM NRL alpha fixed",
                                                                axes_names=["voxel"],
                                                                meta_data=(affine, header))
                outputs["ppm_a_nrl_"+condition_name].meta_data[1]["descrip"] = condition_name

                outputs["ppm_g_nrl_"+condition_name] = xndarray(ppm_g_nrl[:, condition_nb],
                                                                value_label="PPM NRL gamma fixed",
                                                                axes_names=["voxel"],
                                                                meta_data=(affine, header))
                outputs["ppm_g_nrl_"+condition_name].meta_data[1]["descrip"] = condition_name

            if (len(self.contrasts) > 0) and self.computeContrast:
                # keys = list((self.contrasts[nc]) for nc in self.contrasts)
                domContrast = {'contrast': self.contrasts.keys()}
                outputs['contrasts'] = xndarray(contrasts_mean, value_label="Contrast",
                                                axes_names=['voxel', 'contrast'],
                                                axes_domains=domContrast)
                # print 'contrast output:'
                # print outputs['contrasts'].descrip()

                outputs['contrasts_variance'] = xndarray(contrasts_var, value_label="Contrasts_Variance",
                                                         axes_names=['voxel', 'contrast'],
                                                         axes_domains=domContrast)

                outputs['ncontrasts'] = xndarray(contrasts_mean/contrasts_var**.5,
                                                 value_label="Normalized Contrast",
                                                 axes_names=['voxel', 'contrast'],
                                                 axes_domains=domContrast)

                for i, contrast in enumerate(self.contrasts.keys()):
                    header = nibabel.Nifti1Header()
                    outputs["ppm_a_"+contrast] = xndarray(ppm_a_contrasts[:, i],
                                                          value_label="PPM Contrasts alpha fixed",
                                                          axes_names=["voxel"],
                                                          meta_data=(affine, header))
                    outputs["ppm_a_"+contrast].meta_data[1]["descrip"] = contrast

                    outputs["ppm_g_"+contrast] = xndarray(ppm_g_contrasts[:, i],
                                                          value_label="PPM Contrasts gamma fixed",
                                                          axes_names=["voxel"],
                                                          meta_data=(affine, header))
                    outputs["ppm_g_"+contrast].meta_data[1]["descrip"] = contrast

        ################################################################################
        # SIMULATION

        if self.simulation and self.fast:

            labels_vem_audio = roiData.simulation[0]['labels'][0]
            labels_vem_video = roiData.simulation[0]['labels'][1]

            M = labels_proba.shape[0]
            K = labels_proba.shape[1]
            J = labels_proba.shape[2]
            true_labels = np.zeros((K, J))
            true_labels[0, :] = np.reshape(labels_vem_audio, (J))
            true_labels[1, :] = np.reshape(labels_vem_video, (J))
            newlabels = np.reshape(labels_proba[:, 1, :], (M, J))
            se = []
            sp = []
            size = np.prod(labels_proba.shape)

            for i in xrange(0, M):
                se0, sp0, auc = roc_curve(newlabels[i, :].tolist(), true_labels[i, :].tolist())
                se.append(se0)
                sp.append(sp0)
                size = min(size, len(sp0))

            SE = np.zeros((M, size), dtype=float)
            SP = np.zeros((M, size), dtype=float)

            for i in xrange(0, M):
                tmp = np.array(se[i])
                SE[i, :] = tmp[0:size]
                tmp = np.array(sp[i])
                SP[i, :] = tmp[0:size]

            sensData, specData = SE, SP
            axes_names = ['condition', '1-specificity']
            ad = {'1-specificity': specData[0], 'condition': cNames}
            outputs['ROC_audio'] = xndarray(sensData, axes_names=axes_names,
                                            axes_domains=ad,
                                            value_label='sensitivity')

            m = specData[0].min()

            import matplotlib.font_manager as fm
            import matplotlib.pyplot as plt

            plt.figure(200)
            plt.plot(sensData[0], specData[0], '--', color='k', linewidth=2.0, label='m=1')
            plt.hold(True)
            plt.plot(sensData[1], specData[1], color='k', linewidth=2.0, label='m=2')
            # plt.legend(('audio','video'))
            plt.xticks(color='k', size=14, fontweight='bold')
            plt.yticks(color='k', size=14, fontweight='bold')
            # plt.xlabel('1 - Specificity', fontsize=16, fontweight='bold')
            # plt.ylabel('Sensitivity', fontsize=16, fontweight='bold')
            prop = fm.FontProperties(size=14, weight='bold')
            plt.legend(loc=1, prop=prop)
            plt.axis([0., 1., m, 1.02])

            if hasattr(roiData.simulation, 'nrls'):
                true_labels = roiData.simulation.nrls.labels
                true_nrls = roiData.simulation.nrls.data
            elif isinstance(roiData.simulation, dict) and \
                    roiData.simulation.has_key('labels') and \
                    roiData.simulation.has_key('nrls'):
                true_labels = roiData.simulation['labels']
                true_nrls = roiData.simulation['nrls']
            else:
                raise Exception('Simulation can not be retrieved from %s' % str(roiData.simulation))

            domCondition = {'condition': cNames}
            outputs['Truenrls'] = xndarray(true_nrls, value_label="True_nrls",
                                           axes_names=['condition', 'voxel'],
                                           axes_domains=domCondition)
            M = labels_proba.shape[0]
            # K = labels_proba.shape[1]
            J = labels_proba.shape[2]

            newlabels = np.reshape(labels_proba[:, 1, :], (M, J))

            for i in xrange(0, M):
                se0, sp0, auc = roc_curve(newlabels[i, :].tolist(),
                                          true_labels[i, :].tolist())
                se.append(se0)
                sp.append(sp0)
                size = min(size, len(sp0))

            SE = np.zeros((M, size), dtype=float)
            SP = np.zeros((M, size), dtype=float)

            for i in xrange(0, M):
                tmp = np.array(se[i])
                SE[i, :] = tmp[0:size]
                tmp = np.array(sp[i])
                SP[i, :] = tmp[0:size]

        # END SIMULATION
        ##########################################################################
        if self.fast:
            d = {'parcel_size': np.array([nvox])}
            outputs['analysis_duration'] = xndarray(np.array([self.analysis_duration]),
                                                    axes_names=['parcel_size'],
                                                    axes_domains=d)

        return outputs


# Function to use directly in parallel computation
def run_analysis(**params):
    """Function to run the JDE VEM analyzer with parallel computation"""
    # pyhrf.verbose.set_verbosity(1)
    # pyhrf.logger.setLevel(logging.INFO)
    fdata = params.pop('roi_data')
    # print 'doing params:'
    # print params
    vem_analyser = JDEVEMAnalyser(**params)
    return dict([('ROI', fdata.get_roi_id())] + params.items()), vem_analyser.analyse_roi(fdata)



# -*- coding: utf-8 -*-
import logging

from collections import OrderedDict
from pprint import pformat

import numpy as np

from nipy.modalities.fmri.experimental_paradigm import (EventRelatedParadigm,
                                                        BlockParadigm)
from scipy.io.matlab import savemat

from pyhrf.tools import apply_to_leaves
from pyhrf.tools import stack_trees, unstack_trees
from pyhrf.tools._io import load_paradigm_from_csv
from pyhrf.tools._io.spmio import load_paradigm_from_mat


logger = logging.getLogger(__name__)


def merge_onsets(onsets, new_condition, criterion=None, durations=None, discard=None):
    """Convention for definition of onsets or durations.

    .. code::

        OrderedDict({
            'condition_name': [ <array of timings for sess1>,
                                <array of timings for sess2>,
                                ...]
            }

    """

    if discard is None:
        discard = []

    if criterion is None:
        criterion = lambda x: new_condition in x

    new_onsets = onsets.__class__()  # might be dict or OrderedDict
    if durations is not None:
        new_durations = onsets.__class__()

    nb_sessions = len(onsets[onsets.keys()[0]])
    merged_onsets = [np.array([]) for _ in range(nb_sessions)]
    merged_durations = [np.array([]) for _ in range(nb_sessions)]

    for cname, all_ons in onsets.iteritems():
        for isess, ons in enumerate(all_ons):
            if durations is not None:
                dur = durations[cname][isess]
            if criterion(cname):
                merged_onsets[isess] = np.concatenate(
                    (merged_onsets[isess], ons))
                if durations is not None:
                    merged_durations[isess] = np.concatenate(
                        (merged_durations[isess], dur))
            elif cname not in discard:
                if isess == 0:
                    new_onsets[cname] = []
                    if durations is not None:
                        new_durations[cname] = []
                new_onsets[cname].append(ons)
                if durations is not None:
                    new_durations[cname].append(dur)

    for isess in range(nb_sessions):
        sorting = np.argsort(merged_onsets[isess])
        merged_onsets[isess] = merged_onsets[isess][sorting]
        if durations is not None:
            merged_durations[isess] = merged_durations[isess][sorting]

    new_onsets[new_condition] = merged_onsets

    if durations is not None:
        new_durations[new_condition] = merged_durations
        return new_onsets, new_durations
    else:
        return new_onsets

builtin_paradigms = []

onsets_loc = OrderedDict([
    ('calculaudio', [np.array([35.4, 44.7, 48., 83.4, 108., 135.,
                               137.7, 173.7,
                               191.7, 251.7])]),
    ('calculvideo', [np.array([0., 2.4, 125.4, 153., 164.4, 198.,
                               201., 221.4,
                               234., 260.4])]),
    ('clicDaudio', [np.array([11.4, 87., 143.4, 162., 230.])]),
    ('clicDvideo', [np.array([18., 69., 227.7, 275.4, 291.])]),
    ('clicGaudio', [np.array([23.7, 62.4, 170.4, 254.7, 257.4])]),
    ('clicGvideo', [np.array([26.7, 71.4, 116.7, 212.7, 215.7])]),
    ('damier_H', [np.array([8.7, 59.7, 149.4, 176.7, 188.4, 218.7,
                            24.7, 248.4, 266.7, 269.7])]),
    ('damier_V', [np.array([33., 96., 122.7, 140.4, 156., 203.7,
                            207., 210., 264., 278.4])]),
    ('phraseaudio', [np.array([15., 20.7, 29.7, 89.7, 119.4, 146.7,
                               236.7, 284.4,
                               293.4, 296.7])]),
    ('phrasevideo', [np.array([39., 41.7, 56.4, 75., 131.4, 159.,
                               167.7, 195.,
                               246., 288.])]),
])

onsets_loc = OrderedDict([
    ('calculaudio', [np.array([35.4, 44.7, 48., 83.4, 108., 135.,
                               137.7, 173.7,
                               191.7, 251.7])]),
    ('calculvideo', [np.array([0., 2.4, 125.4, 153., 164.4, 198.,
                               201., 221.4,
                               234., 260.4])]),
    ('clicDaudio', [np.array([11.4, 87., 143.4, 162., 230.])]),
    ('clicDvideo', [np.array([18., 69., 227.7, 275.4, 291.])]),
    ('clicGaudio', [np.array([23.7, 62.4, 170.4, 254.7, 257.4])]),
    ('clicGvideo', [np.array([26.7, 71.4, 116.7, 212.7, 215.7])]),
    ('damier_H', [np.array([8.7, 59.7, 149.4, 176.7, 188.4, 218.7,
                            224.7, 248.4, 266.7, 269.7])]),
    ('damier_V', [np.array([33., 96., 122.7, 140.4, 156., 203.7,
                            207., 210., 264., 278.4])]),
    ('phraseaudio', [np.array([15., 20.7, 29.7, 89.7, 119.4, 146.7,
                               236.7, 284.4,
                               293.4, 296.7])]),
    ('phrasevideo', [np.array([39., 41.7, 56.4, 75., 131.4, 159.,
                               167.7, 195.,
                               246., 288.])]),
])

builtin_paradigms.append('loc')

onsets_language_sess1 = dict([
    ('LangForeign', [np.array([24.026, 72.026, 120.026, 144.026])]),
    ('LangForeignRep', [np.array([36.026, 84.026, 132.026, 156.026])]),
    ('LangMat', [np.array([2.60000000e-02, 4.80260000e+01, 9.60260000e+01,
                           1.68026000e+02])]),
    ('LangMatRep', [np.array([12.026, 60.026, 108.026, 180.026])]),
])

durations_language_sess1 = dict([
    ('LangForeign', [np.array([3.021, 2.831, 2.587, 2.325])]),
    ('LangForeignRep', [np.array([3.021, 2.831, 2.587, 2.325])]),
    ('LangMat', [np.array([2.916, 2.753, 2.602, 2.501])]),
    ('LangMatRep', [np.array([2.916, 2.753, 2.602, 2.501])]),
])


onsets_language_sess2 = dict([
    ('LangForeign', [np.array([48.022, 72.022, 96.022, 120.022])]),
    ('LangForeignRep', [np.array([60.022, 84.022, 108.022, 132.022])]),
    ('LangMat', [np.array([2.20000000e-02, 2.40220000e+01, 1.44022000e+02,
                           1.68022000e+02])]),
    ('LangMatRep', [np.array([12.022, 36.022, 156.022, 180.022])]),
])

durations_language_sess2 = dict([
    ('LangForeign', [np.array([2.362, 2.662, 2.643, 2.769])]),
    ('LangForeignRep', [np.array([2.362, 2.662, 2.643, 2.769])]),
    ('LangMat', [np.array([2.839, 2.688, 2.973, 2.774])]),
    ('LangMatRep', [np.array([2.839, 2.688, 2.973, 2.774])]),
])


onsets_language_sess3 = dict([
    ('LangForeign', [np.array([2.90000000e-02, 2.40290000e+01, 7.20290000e+01,
                               1.44029000e+02])]),
    ('LangForeignRep', [np.array([12.029, 36.029, 84.029, 156.029])]),
    ('LangMat', [np.array([48.029, 96.029, 120.029, 168.029])]),
    ('LangMatRep', [np.array([60.029, 108.029, 132.029, 180.029])]),
])

durations_language_sess3 = dict([
    ('LangForeign', [np.array([2.766, 2.625, 2.831, 2.327])]),
    ('LangForeignRep', [np.array([2.766, 2.625, 2.831, 2.327])]),
    ('LangMat', [np.array([2.75, 2.622, 2.847, 2.67])]),
    ('LangMatRep', [np.array([2.75, 2.622, 2.847, 2.67])]),
])


onsets_language_sess4 = dict([
    ('LangForeign', [np.array([3.10000000e-02, 7.20310000e+01, 1.20031000e+02,
                               1.68031000e+02])]),
    ('LangForeignRep', [np.array([12.031, 84.031, 132.031, 180.031])]),
    ('LangMat', [np.array([24.031, 48.031, 96.031, 144.031])]),
    ('LangMatRep', [np.array([36.031, 60.031, 108.031, 156.031])]),
])

durations_language_sess4 = dict([
    ('LangForeign', [np.array([3.292, 2.53, 2.945, 2.775])]),
    ('LangForeignRep', [np.array([3.292, 2.53, 2.945, 2.775])]),
    ('LangMat', [np.array([2.819, 2.771, 2.736, 2.492])]),
    ('LangMatRep', [np.array([2.819, 2.771, 2.736, 2.492])]),
])

builtin_paradigms.extend('language_sess%d' % i for i in range(1, 5))

durations_loc = apply_to_leaves(onsets_loc, lambda x: 2. + np.zeros_like(x))

default_contrasts_loc = OrderedDict({
    "left_click": "clicGaudio + clicGvideo",
    "right_click": "clicDaudio + clicDvideo",
    "checkerboard_H-V": "damier_H - damier_V",
    "checkerboard_V-H": "damier_V - damier_H",
    "left_click-right_click": "clicGaudio-clicDaudio + clicGvideo-clicDvideo",
    "right_click-left_click": "clicDaudio-clicGaudio + clicDvideo-clicGvideo",
    "computation-sentences": "calculaudio-phraseaudio + calculvideo-phrasevideo",
    "sentences-computation": "phraseaudio-calculaudio + phrasevideo-calculvideo",
    "video_computation-sentences": "calculvideo-phrasevideo",
    "audio_computation-sentences": "calculaudio-phraseaudio",
    "motor-cognitive": ("clicGaudio + clicDaudio + clicGvideo + clicDvideo -"
                        " calculaudio - phraseaudio - calculvideo - phrasevideo"),
    "cognitive-motor": ("calculaudio + phraseaudio + calculvideo + phrasevideo -"
                        " clicGaudio - clicDaudio - clicGvideo - clicDvideo"),
    "audio-video": ("calculaudio + phraseaudio + clicGaudio + clicDaudio -"
                    " calculvideo - phrasevideo - clicGvideo - clicDvideo"),
    "video-audio": ("calculvideo + phrasevideo + clicGvideo + clicDvideo -"
                    " calculaudio - phraseaudio - clicGaudio - clicDaudio"),
})


def contrasts_to_spm_vec(condition_list, contrasts):
    res = {}
    cond_d = {}
    for ic, c in enumerate(condition_list):
        cond_d[ic] = c
    for con_name, con in contrasts.iteritems():
        res[con_name] = con
        res[con_name] = res[con_name].replace(' ', '')
        if not con.startswith('+') and not con.startswith('-'):
            res[con_name] = '+' + res[con_name]

        cdict = dict((c, i) for i, c in enumerate(condition_list))

        for c in reversed(sorted(condition_list, key=len)):
            res[con_name] = res[con_name].replace(c, '%02d' % cdict[c])

    for con_name, con in res.iteritems():
        v = np.zeros(len(condition_list))
        for i in range(0, len(con), 3):
            v[int(con[i + 1:i + 3])] = eval('%s1' %
                                            con[i])  # FIXME: don't use eval
        res[con_name] = v
    return res

onsets_un_evnt = OrderedDict([
    ('audio', [np.array([35.])])])
durations_un_evnt = OrderedDict([
    ('audio', [np.array([0.])])])


# audio, video only (video does not comprise damier)
to_discard = ['damier_H', 'damier_V']
o, d = onsets_loc, durations_loc
for c in ['audio', 'video']:
    o, d = merge_onsets(o, c, durations=d, discard=to_discard)
onsets_loc_av, durations_loc_av = o, d
default_contrasts_loc_av = OrderedDict({
    'audio-video': 'audio-video',
    'video-audio': 'video-audio',
})

# audio, video full (video comprises damier)
o, d = onsets_loc, durations_loc
o, d = merge_onsets(o, 'audio', durations=d)
o, d = merge_onsets(o, 'video', criterion=lambda x: 'video' in x or 'damier' in x,
                    durations=d)
onsets_loc_avd, durations_loc_avd = o, d
default_contrasts_loc_avd = OrderedDict({
    'audio-video': 'audio-video',
    'video-audio': 'video-audio',
})


# audio, video, damier
o, d = onsets_loc, durations_loc
o, d = merge_onsets(o, 'audio', durations=d)
o, d = merge_onsets(o, 'video', durations=d)
o, d = merge_onsets(o, 'damier', durations=d)
onsets_loc_av_d, durations_loc_av_d = o, d
default_contrasts_loc_av_d = OrderedDict({
    'audio-video': 'audio-video',
    'video-audio': 'video-audio',
})
builtin_paradigms.append('loc_av_d')

# only one condition (audio)
o, d = onsets_loc, durations_loc
to_discard = ['damier_H',
              'damier_V',
              'clicDvideo',
              'clicGvideo',
              'phrasevideo',
              'calculvideo']
onsets_loc_a, durations_loc_a = merge_onsets(o, 'audio', durations=d,
                                             discard=to_discard)
builtin_paradigms.append('loc_a')

# calcul, phrase only
to_discard = ['damier_H', 'damier_V', 'clicDaudio',
              'clicDvideo', 'clicGvideo', 'clicGaudio']
o, d = onsets_loc, durations_loc
for c in ['calcul', 'phrase']:
    o, d = merge_onsets(o, c, durations=d, discard=to_discard)
onsets_loc_cp_only, durations_loc_cp_only = o, d
default_contrasts_loc_cp_only = OrderedDict({
    "computation-sentences": "calcul - phrase",
    "sentences-computation": "phrase - calcul",
})
builtin_paradigms.append('loc_cp_only')

# calcul only
onsets_loc_c_only = {'calcul': onsets_loc_cp_only['calcul']}
durations_loc_c_only = {'calcul': durations_loc_cp_only['calcul']}
default_contrasts_loc_c_only = OrderedDict()
builtin_paradigms.append('loc_c_only')

# calcul, phrase, clic, damier
o, d = onsets_loc, durations_loc
for c in ['calcul', 'phrase', 'clic', 'damier']:
    o, d = merge_onsets(o, c, durations=d)
onsets_loc_cpcd, durations_loc_cpcd = o, d
default_contrasts_loc_cpcd = OrderedDict({
    "computation-sentences": "calcul - phrase",
    "sentences-computation": "phrase - calcul",
    "motor-cognitive": "calcul + phrase - clic",
    "cognitive-motor": "clic - calcul - phrase",
})
builtin_paradigms.append('loc_cpcd')

onsets_loc_ainsi = OrderedDict([
    ('clicGaudio', [np.array([14.6, 33.3, 89.1, 197.7, 209.0,
                              235.2, 306.8, 344.3, 347.7, 355.9])]),
    ('calculaudio', [np.array([47.9, 59.6, 63.7, 122.7, 153.5,
                               191.0, 194.3, 239.3, 265.5, 340.5])]),
    ('clicGvideo', [np.array([37.1, 104.0, 168.1, 175.6, 217.2,
                              243.1, 280.5, 291.8, 295.5, 359.3])]),
    ('clicDaudio', [np.array([17.9, 127.3, 138.5, 201.4, 224.7,
                              261.4, 284.7, 313.4, 336.4, 363.0])]),
    ('clicDvideo', [np.array([26.2, 44.9, 85.7, 101.0, 288.4,
                              299.3, 310.5, 373.8, 377.6, 397.1])]),
    ('phraseaudio', [np.array([22.5, 29.6, 40.8, 130.6, 171.4,
                               205.6, 321.8, 388.8, 400.1, 404.2])]),
    ('calculvideo', [np.array([0.0, 3.0, 179.0, 213.4, 227.7,
                               273.4, 277.1, 302.7, 318.4, 351.4])]),
    ('phrasevideo', [np.array([52.5, 55.8, 81.6, 108.6, 186.4,
                               221.0, 231.8, 269.7, 333.4, 393.3])]),
])
durations_loc_ainsi = apply_to_leaves(onsets_loc_ainsi, np.zeros_like)

builtin_paradigms.append('loc_ainsi')

# calcul, phrase, clicD, clicG for Ainsi
o, d = onsets_loc_ainsi, durations_loc_ainsi
for c in ['calcul', 'phrase', 'clicD', 'clicG']:
    o, d = merge_onsets(o, c, durations=d)
    onsets_loc_ainsi_cpcd, durations_loc_ainsi_cpcd = o, d
builtin_paradigms.append('loc_ainsi_cpcd')


def check_stim_durations(stim_onsets, stimDurations):
    """ If no durations specified (stimDurations is None or empty np.array)
    then assume spiked stimuli: return a sequence of zeros with same
    shape as onsets sequence.
    Check that durations have same shape as onsets.

    """
    nbc = len(stim_onsets)
    nbs = len(stim_onsets[stim_onsets.keys()[0]])
    if (stimDurations is None or
        (type(stimDurations) == list and
         all([d is None for d in stimDurations]))):
        dur_seq = [[np.array([]) for s in xrange(nbs)] for i in xrange(nbc)]
        stimDurations = OrderedDict(zip(stim_onsets.keys(), dur_seq))

    if stimDurations.keys() != stim_onsets.keys():
        raise Exception('Conditions in stimDurations (%s) differ '
                        'from stim_onsets (%s)' % (stimDurations.keys(),
                                                   stim_onsets.keys()))

    for cn, sdur in stimDurations.iteritems():
        for i, dur in enumerate(sdur):
            if dur is None:
                stimDurations[cn][i] = np.zeros_like(stim_onsets[cn][i])
            elif hasattr(dur, 'len') and len(dur) == 0:
                stimDurations[cn][i] = np.zeros_like(stim_onsets[cn][i])
            elif hasattr(dur, 'size') and dur.size == 0:
                stimDurations[cn][i] = np.zeros_like(stim_onsets[cn][i])
            else:
                if not isinstance(stimDurations, np.ndarray):
                    stimDurations[cn][i] = np.array(dur)

                assert len(stimDurations[cn][i]) == len(stim_onsets[cn][i])

    return stimDurations


def extend_sampled_events(sampled_events, sampled_durations):
    """ Add events to encode stimulus duration
    """
    if 0:
        print 'sampled_events:', len(sampled_events)
        print sampled_events
        print 'sampled_durations:', len(sampled_durations)
        print sampled_durations
    extended_events = set(sampled_events)
    for io, o in enumerate(sampled_events):
        extended_events.update(range(o + 1, o + sampled_durations[io]))

    return np.array(sorted(list(extended_events)), dtype=int)


def restarize_events(events, durations, dt, t_max):
    """ build a binary sequence of events. Each event start is approximated
    to the nearest time point on the time grid defined by dt and t_max.
    """
    smpl_events = np.array(np.round_(np.divide(events, dt)), dtype=int)
    smpl_durations = np.array(np.round_(np.divide(durations, dt)), dtype=int)
    smpl_events = extend_sampled_events(smpl_events, smpl_durations)
    logger.debug("sampled_onsets: %s", str(smpl_events.shape))
    logger.debug(smpl_events)
    if np.allclose(t_max % dt, 0):
        bin_seq = np.zeros(int(t_max / dt) + 1)
    else:
        bin_seq = np.zeros(int(np.round((t_max + dt) / dt)))
    bin_seq[smpl_events] = 1
    logger.debug('bins :')
    logger.debug(bin_seq)

    return bin_seq


class Paradigm:
    """
    """

    def __init__(self, stimOnsets, sessionDurations=None, stimDurations=None):
        """

        Args:
            *stimOnsets* (dict of list) :
                dictionary mapping a condition name to a list of session
                stimulus time arrivals.
                eg:
                {'cond1' : [<session 1 onsets>, <session 2 onsets>]
                 'cond2' : [<session 1 onsets>, <session 2 onsets>]
                 }
            *sessionDurations* (1D numpy float array): durations for all sessions
            *stimDurations* (dict of list) : same structure as stimOnsets.
                             If None, spiked stimuli are assumed (ie duration=0).

        """

        # print 'stimOnsets;', stimOnsets.__class__
        # print 'stimDurations:', stimDurations
        assert isinstance(stimOnsets, dict)
        c = stimOnsets.keys()[0]
        assert isinstance(stimOnsets[c], list)
        # print stimOnsets[c][0]
        assert isinstance(stimOnsets[c][0], np.ndarray)
        assert stimOnsets[c][0].ndim == 1

        if sessionDurations is not None:
            assert isinstance(sessionDurations, np.ndarray) or \
                isinstance(sessionDurations, list)
            assert len(sessionDurations) == len(stimOnsets[c])

        self.stimOnsets = stimOnsets
        # for k,v in stimOnsets.items():
        #     for isess,vsess in enumerate(v):
        #         if not isinstance(vsess, np.ndarray):
        #             self.stimOnsets[k][isess] = np.array(vsess)

        self.stimDurations = check_stim_durations(stimOnsets, stimDurations)

        self.nbSessions = len(self.stimOnsets[self.stimOnsets.keys()[0]])
        if sessionDurations is None:
            # print 'compute sessionDurations ...'
            sessionDurations = []
            max_onset = 0.
            for s in xrange(self.nbSessions):
                for cn, v in stimOnsets.items():
                    argmax_onset = v[s].argmax()
                    if v[s][argmax_onset] > max_onset:
                        max_onset = v[s][argmax_onset]
                        max_durations = self.stimDurations[cn][s][argmax_onset]
                sessionDurations.append(max_onset + max_durations)

            # sessionDurations = [max([max(v[s]) for v in stimOnsets.values()]) \
                # for s in xrange(self.nbSessions)]
        self.sessionDurations = sessionDurations

    def delete_condition(self, cond):
        self.stimOnsets.pop(cond, None)
        self.stimDurations.pop(cond, None)

    @classmethod
    def from_session_dict(self, d, sessionDurations=None):
        nd = stack_trees([d[sess] for sess in sorted(d.keys())])
        return Paradigm(nd['onsets'], sessionDurations, nd['stimulusLength'])

    def __repr__(self):
        s = 'Paradigm('
        s += 'stimOnsets=%s,' % repr(self.stimOnsets)
        s += 'sessionDurations=%s,' % repr(self.sessionDurations)
        s += 'stimDurations=%s)' % repr(self.stimDurations)
        return s

    @classmethod
    def from_csv(self, csvFile, delim=None):
        """
        Create a Paradigm object from a CSV file which columns are:
        session, task name, stimulation onset, stimulation duration, [amplitude]
        """
        onsets, durations = load_paradigm_from_csv(csvFile, delim)
        return Paradigm(onsets, stimDurations=durations)

    @classmethod
    def from_spm_mat(self, spm_mat_file):
        """
        TODO: handle session durations
        """
        par_data, _ = load_paradigm_from_mat(spm_mat_file)
        par_data = stack_trees([par_data[s] for s in sorted(par_data.keys())])
        return Paradigm(par_data['onsets'],
                        stimDurations=par_data['stimulusLength'])

    def save_spm_mat_for_1st_level_glm(self, mat_file, session=0):
        ordered_names = sorted(self.stimOnsets.keys())
        to_save = {
            'onsets': np.array([self.stimOnsets[n][session]
                                for n in ordered_names],
                               dtype=object),
            'names': np.array(ordered_names, dtype=object),
            'durations': np.array([self.stimDurations[n][session]
                                   for n in ordered_names], dtype=object),
        }

        # print 'to_save:'
        # print to_save
        # print 'stimOnsets:', self.stimOnsets
        savemat(mat_file, to_save, oned_as='row')

    def join_sessions(self):
        if self.nbSessions == 1:
            return self

        ons = OrderedDict((n, [o])
                          for n, o in self.get_joined_onsets().iteritems())
        dur = OrderedDict((n, [d])
                          for n, d in self.get_joined_durations().iteritems())
        return Paradigm(ons, self.sessionDurations[-1:], dur)

    def to_nipy_paradigm(self):

        p = self.join_sessions()
        sorted_conds = sorted(p.stimOnsets.keys())
        onsets = unstack_trees(p.stimOnsets)
        durations = unstack_trees(p.stimDurations)

        cond_ids = [np.hstack([[c] * len(o)
                               for c, o in sorted(sess_ons.items())])
                    for sess_ons in onsets]
        # print 'cond_ids:', cond_ids
        onsets = [np.hstack([sess_ons[c] for c in sorted_conds])
                  for sess_ons in onsets]
        durations = [np.hstack([sess_dur[c] for c in sorted_conds])
                     for sess_dur in durations]
        # print 'onsets:'
        # print onsets
        # print 'durations'
        # print durations
        # print '[(dur>0.).any() for dur in durations]:'
        # print [(dur>0.).any() for dur in durations]

        if any([(dur > 0.).any() for dur in durations]):
            # Block paradigm
            if len(onsets) > 1:
                dd = [('session%02d' % i, BlockParadigm(d[0], d[1], d[2]))
                      for i, d in enumerate(zip(cond_ids, onsets, durations))]
                return dict(dd)
            else:
                return BlockParadigm(cond_ids[0], onsets[0], durations[0],
                                     amplitude=None)
        else:
            if len(onsets) > 1:

                dd = [('session%02d' % i, EventRelatedParadigm(d[0], d[1]))
                      for i, d in enumerate(zip(cond_ids, onsets))]
                return dict(dd)
            else:
                return EventRelatedParadigm(cond_ids[0], onsets[0],
                                            amplitude=None)

    def save_csv(self, csvFile):

        s = ''
        for cn, sessOns in self.stimOnsets.iteritems():
            sessDur = self.stimDurations[cn]
            for iSess, ons in enumerate(sessOns):
                durs = sessDur[iSess]
                for on, dur in zip(ons, durs):
                    s += '%d "%s" %f %f 1.\n' % (iSess, cn, on, dur)
        f = open(csvFile, 'w')
        f.write(s)
        f.close()

    def get_info(self, long=True):
        s = ''
        s += 'sessionDurations: %s\n' % str(self.sessionDurations)
        allOns = []
        for oc in self.stimOnsets.values():  # parse conditions
            for o in oc:  # parse sessions
                allOns.extend(o)

        allOns = np.array(allOns)
        allOns.sort()
        last = allOns.max()
        meanISI = np.diff(allOns).mean()
        stdISI = np.diff(allOns).std()
        s += ' - onsets : ISI=%1.2f(%1.2f)sec' \
             ' - last event: %1.2fsec\n' % (meanISI, stdISI, last)
        for stimName in self.stimOnsets.keys():
            ntrials = [len(o) for o in self.stimOnsets[stimName]]
            s += '     >> %s, trials per session: %s\n' % (
                stimName, str(ntrials))
            if long:
                s += pformat(self.stimOnsets[stimName], indent=5)
                s += pformat(self.stimDurations[stimName], indent=5)
        return s

    def get_joined_onsets(self):
        """ For each condition, join onsets of all sessions.
        """
        jOnsets = OrderedDict([])
        for cn, sessOns in self.stimOnsets.iteritems():
            scanStartTime = 0.
            jons = np.array([])
            for iSess, ons in enumerate(sessOns):
                shiftedOns = ons + scanStartTime
                jons = np.concatenate((jons, shiftedOns))
                scanStartTime += self.sessionDurations[iSess]
            jOnsets[cn] = jons
        return jOnsets

    def get_joined_onsets_dim(self):
        """ For each condition, join onsets of all sessions.
        """
        jOnsets = OrderedDict([])
        for cn, sessOns in self.stimOnsets.iteritems():
            jOnsets[cn] = np.array(sessOns)
        return jOnsets

    def get_nb_trials(self):
        pass

    def get_joined_durations(self):
        """ For each condition, join stimulus durations of all sessions.
        """
        jDurations = OrderedDict([])
        for cn, sessDur in self.stimDurations.iteritems():
            jDurations[cn] = np.concatenate(tuple(sessDur))
        return jDurations

    def get_joined_durations_dim(self):
        """ For each condition, join stimulus durations of all sessions.
        """
        jDurations = OrderedDict([])
        for cn, sessDur in self.stimDurations.iteritems():
            jDurations[cn] = np.array(sessDur)
        return jDurations

    def get_stimulus_names(self):
        return self.stimOnsets.keys()

    def get_t_max(self):
        ns = len(self.sessionDurations)
        return max([self.sessionDurations[i] for i in xrange(ns)])

    def get_rastered(self, dt, tMax=None):
        """ Return binary sequences of stimulus arrivals. Each stimulus event is approximated to the closest time point
        on the time grid defined by dt. eg return ::

            { 'cond1' : [np.array([ 0 0 0 1 0 0 1 1 1 0 1]),
                         np.array([ 0 1 1 1 0 0 1 0 1 0 0])] },
              'cond2' : [np.array([ 0 0 0 1 0 0 1 1 1 0 0]),
                         np.array([ 1 1 0 1 0 1 0 0 0 0 0])] },

        Parameters
        ----------
        dt : float
            temporal resolution of the target grid
        tMax : float
            total duration of the paradigm. If None, then use the session lengths

        """
        assert dt > 0.
        rasteredParadigm = OrderedDict({})
        # print 'self.stimOnsets:'
        # print self.stimOnsets
        if tMax is None:
            tMax = self.get_t_max()
        for cn in self.stimOnsets.iterkeys():
            par = []
            for iSess, ons in enumerate(self.stimOnsets[cn]):
                dur = self.stimDurations[cn][iSess]
                if 0:
                    print '-----------------'
                    print 'dur:', dur
                    print 'ons:', ons
                    print 'tMax:', tMax
                    print 'dt:', dt
                    print '-----------------'
                binaryEvents = restarize_events(ons, dur, dt, tMax)
                par.append(binaryEvents)
            rasteredParadigm[cn] = np.vstack(par)

        return rasteredParadigm

    def get_joined_and_rastered(self, dt):
        rpar = self.get_rastered(dt)
        for cn in rpar.iterkeys():
            rpar[cn] = np.concatenate(rpar[cn])

        return rpar



# -*- coding: utf-8 -*-
import os
import os.path as op
import logging

import numpy as np

from pyhrf.ui.treatment import FMRITreatment
from pyhrf.ui.jde import JDEMCMCAnalyser


logger = logging.getLogger(__name__)


def jde_analyse_2steps_v1(output_dir, fmri_data, dt, nb_iterations, brf_var=None,
                          do_sampling_brf_var=False, prf_var=None,
                          do_sampling_prf_var=False):
    """
    #Return:
    #    dict of outputs
    """
    nb_iterations_m1 = nb_iterations_m2 = nb_iterations

    jde_output_dir = op.join(output_dir, 'bold_only')
    if not op.exists(jde_output_dir):
        os.makedirs(jde_output_dir)

    dummy_sampler = dummy_jde(fmri_data, dt)

    prf_ini_m1 = np.zeros_like(dummy_sampler.get_variable('prf').finalValue)
    prls_ini_m1 = np.zeros_like(dummy_sampler.get_variable('prl').finalValue)
    jde_mcmc_sampler_m1 = physio_build_jde_mcmc_sampler(
        nb_iterations_m1, 'basic_regularized', prf_ini=prf_ini_m1,
        do_sampling_prf=False, prls_ini=prls_ini_m1, do_sampling_prls=False,
        brf_var_ini=brf_var, do_sampling_brf_var=do_sampling_brf_var,
        prf_var_ini=brf_var, do_sampling_prf_var=False, flag_zc = False)

    logger.info('JDE first pass -> BOLD fit')
    analyser_m1 = JDEMCMCAnalyser(jde_mcmc_sampler_m1, copy_sampler=False,
                                  dt=dt)
    analyser_m1.set_pass_errors(False)
    tjde_mcmc_m1 = FMRITreatment(fmri_data, analyser_m1,
                                 output_dir=jde_output_dir)
    outputs_m1, fns_m1 = tjde_mcmc_m1.run()

    jde_output_dir = op.join(output_dir, 'perf_only_from_res')
    if not op.exists(jde_output_dir):
        os.makedirs(jde_output_dir)

    brf_m1 = jde_mcmc_sampler_m1.get_variable('brf').finalValue
    omega = jde_mcmc_sampler_m1.get_variable('prf').omega_operator
    print omega.shape
    print np.concatenate(([0],[0],[0],brf_m1,[0],[0],[0])).shape
    #prf_m2 = np.dot(omega, brf_m1)
    prf_m2 = np.dot(omega, np.concatenate(([0],[0],[0],brf_m1,[0],[0],[0])))[3:-3]

    if 0:
        import matplotlib.pyplot as plt
        plt.plot(brf_m1)
        plt.plot(prf_m2)
        plt.show()

    if 0:
        import matplotlib.pyplot as plt
        f, axarr = plt.subplots(2, 2)
        axarr[0, 0].matshow(omega)
        axarr[0, 1].plot(prf_m2)
        import copy
        omega2 = copy.deepcopy(omega)
        omega2[-5:,:5] = 0.
        omega2[:5,-5:] = 0.
        prf_m2 = np.dot(omega2, brf_m1)
        axarr[1, 0].matshow(omega2)
        axarr[1, 1].plot(prf_m2)
        plt.show()

    #print 'PRF M2 shape = ', prf_m2.shape
    #force begin & end to be zero
    #XXX TODO: rather fix omega to avoid this kind of thing or use zero-contrainst again
    #prf_m2[0] = 0.
    #prf_m2[-5:] = 0.
    brls_m1 = jde_mcmc_sampler_m1.get_variable('brl').finalValue
    labels_m1 = jde_mcmc_sampler_m1.get_variable('label').finalValue >= 0.5
    labels_m1 = labels_m1.astype(np.int32)
    drift_m1 = jde_mcmc_sampler_m1.get_variable('drift_coeff').finalValue
    drift_var_m1 = jde_mcmc_sampler_m1.get_variable('drift_var').finalValue
    noise_var_m1 = jde_mcmc_sampler_m1.get_variable('noise_var').finalValue

    logger.info('ASL JDE second pass -> Perfusion fit')
    logger.info('Physiological prior stochastic regularized')

    jde_mcmc_sampler_m2 = \
      physio_build_jde_mcmc_sampler(nb_iterations_m2, 'physio_stochastic_not_regularized',
                                    #prf_ini=prf_m2,
                                    do_sampling_prf=True,
                                    flag_zc=True,
                                    prf_var_ini = prf_var,
                                    do_sampling_prf_var = do_sampling_prf_var,
                                    brls_ini=brls_m1, #if prls_ini_m1, it should be the same
                                    do_sampling_brls=False,
                                    brf_ini=brf_m1, #if prf_ini_m1, PRF sampler 'basic_regularized'
                                    do_sampling_brf=False,
                                    brf_var_ini=brf_var,
                                    do_sampling_brf_var=False,
                                    #perf_bl_ini=perf_bl_m1,
                                    drift_ini=drift_m1,
                                    do_sampling_drift = False,
                                    drift_var_ini = drift_var_m1,
                                    do_sampling_drift_var = False,
                                    noise_var_ini=noise_var_m1,
                                    labels_ini=labels_m1,
                                    do_sampling_labels=True)

    """ available_priors = ['physio_stochastic_regularized',
                            'physio_stochastic_not_regularized',
                            'physio_deterministic',
                            'physio_deterministic_hack',
                            'basic_regularized']                """

    analyser_m2 = JDEMCMCAnalyser(jde_mcmc_sampler_m2, dt=dt)
    analyser_m2.set_pass_errors(False)
    tjde_mcmc_m2 = FMRITreatment(fmri_data, analyser_m2,
                                 output_dir=jde_output_dir)
    outputs_m2, fns_m2 = tjde_mcmc_m2.run()

    # combine outputs
    outputs = {}
    for o in ['brf_pm', 'brl_pm']:
        outputs[o] = outputs_m1[o]
    outputs['noise_var_pm_m1'] = outputs_m1['noise_var_pm']
    for o in ['prf_pm', 'prl_pm', 'label_pm']:
        outputs[o] = outputs_m2[o]
    outputs['noise_var_pm_m2'] = outputs_m2['noise_var_pm']

    return outputs_m1


"""
def jde_analyse_2steps(output_dir, fmri_data, dt, nb_iterations, brf_var=None,
                       do_sampling_brf_var=False):


    #Return:
    #    dict of outputs


    nb_iterations_m1 = nb_iterations_m2 = nb_iterations

    jde_output_dir = op.join(output_dir, 'bold_only')
    if not op.exists(jde_output_dir):
        os.makedirs(jde_output_dir)

    dummy_sampler = dummy_jde(fmri_data, dt)
    prf_ini_m1 = np.zeros_like(dummy_sampler.get_variable('prf').finalValue)
    prls_ini_m1 = np.zeros_like(dummy_sampler.get_variable('prl').finalValue)

    jde_mcmc_sampler_m1 = \
      physio_build_jde_mcmc_sampler(nb_iterations_m1, 'basic_regularized',
                                    prf_ini=prf_ini_m1,
                                    do_sampling_prf=False,
                                    prls_ini=prls_ini_m1,
                                    do_sampling_prls=False,
                                    brf_var_ini=brf_var,
                                    do_sampling_brf_var=do_sampling_brf_var,
                                    prf_var_ini=brf_var,
                                    do_sampling_prf_var=False)

    logger.info('JDE first pass -> BOLD fit')
    analyser_m1 = JDEMCMCAnalyser(jde_mcmc_sampler_m1, copy_sampler=False,
                                  dt=dt)
    analyser_m1.set_pass_errors(False)
    tjde_mcmc_m1 = FMRITreatment(fmri_data, analyser_m1,
                                 output_dir=jde_output_dir)
    outputs_m1, fns_m1 = tjde_mcmc_m1.run()


    jde_output_dir = op.join(output_dir, 'perf_only_from_res')
    if not op.exists(jde_output_dir):
        os.makedirs(jde_output_dir)


    brf_m1 = jde_mcmc_sampler_m1.get_variable('brf').finalValue
    prf_m2 = np.dot(jde_mcmc_sampler_m1.get_variable('prf').omega_operator,
                    brf_m1)

    #force begin & end to be zero
    #TODO: rather fix omega to avoid this kind of thing or
    # use zero-contrainst again
    #prf_m2[0] = 0.
    #prf_m2[-5:] = 0.

    brls_m1 = jde_mcmc_sampler_m1.get_variable('brl').finalValue
    labels_m1 = jde_mcmc_sampler_m1.get_variable('label').finalValue >= 0.5
    labels_m1 = labels_m1.astype(np.int32)

    perf_bl_m1 = jde_mcmc_sampler_m1.get_variable('perf_baseline').finalValue
    drift_m1 = jde_mcmc_sampler_m1.get_variable('drift_coeff').finalValue
    drift_var_m1 = jde_mcmc_sampler_m1.get_variable('drift_var').finalValue
    noise_var_m1 = jde_mcmc_sampler_m1.get_variable('noise_var').finalValue

    logger.info('ASL JDE second pass -> Perfusion fit')
    jde_mcmc_sampler_m2 = \
      physio_build_jde_mcmc_sampler(nb_iterations_m2, 'basic_regularized',
                                    flag_zc = True,
                                    prf_ini=prf_m2,
                                    do_sampling_prf=False,
                                    prf_var_ini=brf_var,
                                    do_sampling_prf_var=False,
                                    brls_ini=brls_m1,
                                    do_sampling_brls=False,
                                    brf_ini=brf_m1,
                                    do_sampling_brf=False,
                                    brf_var_ini=brf_var,
                                    do_sampling_brf_var=False,
                                    perf_bl_ini=perf_bl_m1,
                                    drift_ini=drift_m1,
                                    do_sampling_drift = False,
                                    drift_var_ini = drift_var_m1,
                                    do_sampling_drift_var = False,
                                    noise_var_ini=noise_var_m1,
                                    labels_ini=labels_m1,
                                    do_sampling_labels=True)

    analyser_m2 = JDEMCMCAnalyser(jde_mcmc_sampler_m2, dt=dt)
    analyser_m2.set_pass_errors(False)
    tjde_mcmc_m2 = FMRITreatment(fmri_data, analyser_m2,
                                 output_dir=jde_output_dir)
    outputs_m2, fns_m2 = tjde_mcmc_m2.run()

    # combine outputs
    outputs = {}
    for o in ['brf_pm', 'brl_pm']:
        outputs[o] = outputs_m1[o]

    outputs['noise_var_pm_m1'] = outputs_m1['noise_var_pm']

    for o in ['prf_pm', 'prl_pm', 'label_pm']:
        outputs[o] = outputs_m2[o]
    outputs['noise_var_pm_m2'] = outputs_m2['noise_var_pm']

    return outputs
"""

def dummy_jde(fmri_data, dt):
    print 'run dummy_jde ...'
    jde_mcmc_sampler = \
        physio_build_jde_mcmc_sampler(3, 'basic_regularized',
                                        do_sampling_prf=False,
                                        do_sampling_brf=False,
                                        do_sampling_prls=False,
                                        do_sampling_labels=False,
                                        do_sampling_prf_var=False,
                                        do_sampling_brf_var=False,
                                        brf_var_ini=np.array([0.1]),
                                        prf_var_ini=np.array([0.1]))

    analyser = JDEMCMCAnalyser(jde_mcmc_sampler, copy_sampler=False, dt=dt)
    analyser.set_pass_errors(False)
    tjde_mcmc = FMRITreatment(fmri_data, analyser, output_dir=None)
    outputs, fns = tjde_mcmc.run()
    print 'dummy_jde done!'
    return tjde_mcmc.analyser.sampler



def physio_build_jde_mcmc_sampler(nb_iterations,
                                  rf_prior,
                                  flag_zc = False,
                                  brf_var_ini=None,
                                  prf_var_ini=None,
                                  do_sampling_brf_var=False,
                                  do_sampling_prf_var=False,
                                  prf_ini=None,
                                  do_sampling_prf=True,
                                  prls_ini=None,
                                  do_sampling_prls=True,
                                  brf_ini=None,
                                  do_sampling_brf=True,
                                  brls_ini=None,
                                  do_sampling_brls=True,
                                  perf_bl_ini=None,
                                  do_sampling_perf_bl=True,
                                  do_sampling_perf_var=True,
                                  drift_ini=None,
                                  do_sampling_drift = True,
                                  drift_var_ini=None,
                                  do_sampling_drift_var = True,
                                  noise_var_ini=None,
                                  labels_ini=None,
                                  do_sampling_labels=True,
                                  ):
    """
    """
    #import pyhrf.jde.asl_physio as jap
    import pyhrf.jde.asl_physio_1step as jap
    #print 'brf_var_ini', brf_var_ini
    #print 'prf_var_ini', prf_var_ini
    sampler_params = {
            'nb_iterations' : nb_iterations,
            'smpl_hist_pace' : -1,
            'obs_hist_pace' : -1,
            'brf' : \
            jap.PhysioBOLDResponseSampler(val_ini=brf_ini,
                                          zero_constraint=flag_zc,
                                          do_sampling=do_sampling_brf,
                                          normalise = 0.,
                                          use_true_value=False),
            'brf_var' : \
                jap.PhysioBOLDResponseVarianceSampler(\
                    val_ini=np.array([brf_var_ini]),
                    do_sampling=do_sampling_brf_var),
            'prf' : \
                jap.PhysioPerfResponseSampler(val_ini=prf_ini,
                                              zero_constraint=flag_zc,
                                              do_sampling=do_sampling_prf,
                                              normalise = 0.,
                                              smooth_order = 2,
                                              use_true_value=False,
                                              prior_type=rf_prior),
            'prf_var' : \
                jap.PhysioPerfResponseVarianceSampler(\
                    val_ini=np.array([prf_var_ini]),
                    do_sampling=do_sampling_prf_var),
            'noise_var' : \
                jap.NoiseVarianceSampler(val_ini=noise_var_ini,
                                         use_true_value=False,
                                         do_sampling=True),
            'drift_var' : \
                jap.DriftVarianceSampler(use_true_value=False,
                                         do_sampling=True),
            'drift' : \
                jap.DriftCoeffSampler(val_ini=drift_ini,
                                      use_true_value=False,
                                      do_sampling=True),
            'bold_response_levels' : \
                jap.BOLDResponseLevelSampler(val_ini=brls_ini,
                                             use_true_value=False,
                                             do_sampling=do_sampling_brls),
            'perf_response_levels' : \
                jap.PerfResponseLevelSampler(val_ini=prls_ini,
                                             use_true_value=False,
                                             do_sampling=do_sampling_prls),
            'bold_mixt_params' : \
                jap.BOLDMixtureSampler(use_true_value=False,
                                       do_sampling=do_sampling_brls),
            'perf_mixt_params' : \
                jap.PerfMixtureSampler(use_true_value=False,
                                       do_sampling=do_sampling_prls),
            'labels' : \
                jap.LabelSampler(val_ini=labels_ini,
                                 use_true_value=False,
                                 do_sampling=do_sampling_labels),
            'perf_baseline' : \
                jap.PerfBaselineSampler(val_ini=perf_bl_ini,
                                        use_true_value=False,
                                        do_sampling=do_sampling_perf_bl),
            'perf_baseline_var' : \
                jap.PerfBaselineVarianceSampler(use_true_value=False,
                                                do_sampling=do_sampling_perf_var),
            'check_final_value' : 'none',
            'output_fit': True,
        }
    sampler = jap.ASLPhysioSampler(**sampler_params)
    return sampler




"""
Performs parcellation to a list of subjects

Directory structure:

- subject:

    * preprocessed_data --> GM+WM mask, functional data, normalised tissue masks
    * t_maps --> T-maps previously computed with GLM (nipy, SALMA)
    * parcellation --> Output

"""

import os
import os.path as op
import numpy as np
from glob import glob


def make_parcellation(subject, dest_dir='parcellation', roi_mask_file=None):
    """
    Perform a functional parcellation from input fmri data
    
    Return: parcellation file name (str)
    """
    # Loading names for folders and files
    # - T maps (input)
    #func_files = glob(op.join(op.join(op.join('./', subject), \
    #                    't_maps'), 'BOLD*nii'))
    #func_files = glob(op.join('./', subject, 'ASLf', 'spm_analysis', \
    #                            'Tmaps*img'))
    func_files = glob(op.join('./', subject, 'ASLf', 'spm_analysis', \
                                'spmT*img'))
    print 'Tmap files: ', func_files 

    # - Mask (input)
    #spm_mask_file = op.join(spm_maps_dir, 'mask.img')
    mask_dir = op.join('./', subject, 'preprocessed_data')
    if not op.exists(mask_dir): os.makedirs(mask_dir)
    mask_file = op.join(mask_dir, 'mask.nii')
    mask = op.join(mask_dir, 'rcut_tissue_mask.nii')
    volume = op.join('./', subject, 'ASLf', 'funct', 'coregister', \
                     'mean' + subject + '_ASLf_correctionT1_0001.nii')
    make_mask(mask, volume, mask_file)

    # - parcellation (output)
    parcellation_dir = op.join('./', subject, dest_dir)
    if not op.exists(parcellation_dir): os.makedirs(parcellation_dir)
    pfile = op.join(parcellation_dir, 'parcellation_func.nii')

    # Parcellation
    from pyhrf.parcellation import make_parcellation_from_files 
    #make_parcellation_from_files(func_files, mask_file, pfile, 
    #                             nparcels=200, method='ward_and_gkm')

    # Masking with a ROI so we just consider parcels inside 
    # a certain area of the brain
    #if roi_mask_file is not None:   
    if 0: #for ip in np.array([11, 51, 131, 194]):
        #ip = 200

        #print 'Masking parcellation with roi_mask_file: ', roi_mask_file
        print 'Masking ROI: ', ip
        pfile_masked = op.join(parcellation_dir, 'parcellation_func_masked_roi' + str(ip) + '.nii')

        from pyhrf.ndarray import xndarray
        parcellation = xndarray.load(pfile)
        #m = xndarray.load(roi_mask_file)
        #parcels_to_keep = np.unique(parcellation.data * m.data)
        masked_parcellation = xndarray.xndarray_like(parcellation)
        #for ip in parcels_to_keep:
        #    masked_parcellation.data[np.where(parcellation.data==ip)] = ip
        
        masked_parcellation.data[np.where(parcellation.data==ip)] = ip
        masked_parcellation.save(pfile_masked)

    from pyhrf.ndarray import xndarray
    for tmap in func_files:
        func_file_i = xndarray.load(tmap)
        func_data = func_file_i.data
        #func_file_i.data[np.where(func_data<0.1*func_data.max())] = 0

        parcellation = xndarray.load(pfile)
        print func_data.max()
        print func_data.argmax()
        #ip = parcellation.data[func_data.argmax()]
        ip = parcellation.data[np.unravel_index(func_data.argmax(), parcellation.data.shape)]
        print ip.shape

        masked_parcellation = xndarray.xndarray_like(parcellation)
        print masked_parcellation.data.shape
        print parcellation.data.shape
        masked_parcellation.data[np.where(parcellation.data==ip)] = ip
        masked_parcellation.save(tmap[:-4] + '_parcelmax.nii')

    return pfile


def make_mask(mask, volume, mask_file):
    from pyhrf.ndarray import xndarray
    m = xndarray.load(mask)
    v = xndarray.load(volume)
    
    mask = m.copy()
    #mask.data = np.zeros_like(m.data)
    mask.data[np.where(v.data == 0.)] = 0
    mask.save(mask_file)
    return 


if __name__ == '__main__':
    
    #subjects = ['RG130377', 'CD110147']
    subjects = ['AINSI_001_GC', 'AINSI_005_SB', 'AINSI_010_TV']
    #subjects = ['AINSI_010_TV']

    for subject in subjects:
        
        print 'Treating subject:', subject
        pfile = make_parcellation(subject)

        print 'parcellation result in file ', pfile




# -*- coding: utf-8 -*-
"""VEM BOLD Constrained

File that contains function for BOLD data analysis with positivity
and l2-norm=1 constraints.

It imports functions from vem_tools.py in pyhrf/vbjde
"""

import time
import copy
import logging
import os

import os.path as op
import numpy as np

import pyhrf
import pyhrf.vbjde.vem_tools as vt

from pyhrf.boldsynth.hrf import getCanoHRF, genGaussianSmoothHRF
from pyhrf.sandbox.physio_params import (PHY_PARAMS_KHALIDOV11,
                                         linear_rf_operator,
                                         create_physio_brf,
                                         create_physio_prf)


import matplotlib
import matplotlib.pyplot as plt
try:
    os.environ["DISPLAY"]
except KeyError:
    matplotlib.use("Agg")
    plt.switch_backend("Agg")
else:
    try:
        matplotlib.use("Qt4Agg")
        plt.switch_backend("Qt4Agg")
    except ImportError:
        pass


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

eps = 1e-6

#@profile
def Main_vbjde_physio(graph, Y, Onsets, durations, Thrf, K, TR, beta, dt,
                      scale=1, estimateSigmaH=True, estimateSigmaG=True,
                      sigmaH=0.05, sigmaG=0.05, gamma_h=0, gamma_g=0,
                      NitMax=-1, NitMin=1, estimateBeta=True, PLOT=False,
                      contrasts=[], computeContrast=False,
                      idx_first_tag=0, simulation=None, sigmaMu=None,
                      estimateH=True, estimateG=True, estimateA=True,
                      estimateC=True, estimateZ=True, estimateNoise=True,
                      estimateMP=True, estimateLA=True, use_hyperprior=False,
                      positivity=False, constraint=False,
                      phy_params=PHY_PARAMS_KHALIDOV11, prior='omega'):

    logger.info("EM for ASL!")
    np.random.seed(6537540)
    logger.info("data shape: ")
    logger.info(Y.shape)

    Thresh = 1e-5
    D, M = np.int(np.ceil(Thrf / dt)) + 1, len(Onsets)
    #D, M = np.int(np.ceil(Thrf / dt)), len(Onsets)
    N, J = Y.shape[0], Y.shape[1]
    Crit_AH, Crit_CG, cTime, rerror, FE = 1, 1, [], [], []
    EP, EPlh, Ent = [],[],[]
    Crit_H, Crit_G, Crit_Z, Crit_A, Crit_C = 1, 1, 1, 1, 1
    cAH, cCG, AH1, CG1 = [], [], [], []
    cA, cC, cH, cG, cZ = [], [], [], [], []
    h_norm, g_norm = [], []
    SUM_q_Z = [[] for m in xrange(M)]
    mua1 = [[] for m in xrange(M)]
    muc1 = [[] for m in xrange(M)]
    AH1, CG1 = np.zeros((J, M, D)), np.zeros((J, M, D))
    sigmaH = sigmaH * J / 100
    print sigmaH
    gamma_h = gamma_h * 100 / J
    print gamma_h

    # Beta data
    MaxItGrad = 200
    gradientStep = 0.005
    gamma = 7.5
    print 'gamma = ', gamma
    print 'voxels = ', J
    maxNeighbours, neighboursIndexes = vt.create_neighbours(graph, J)

    # Conditions
    print 'durations'
    print durations
    X, XX, condition_names = vt.create_conditions_block(Onsets, durations, M, N, D, TR, dt)
    #X, XX, condition_names = vt.create_conditions(Onsets, M, N, D, TR, dt)

    # Covariance matrix
    #R = vt.covariance_matrix(2, D, dt)
    _, R_inv = genGaussianSmoothHRF(False, D, dt, 1., 2)
    R = np.linalg.inv(R_inv)
    # Noise matrix
    Gamma = np.identity(N)
    # Noise initialization
    sigma_eps = np.ones(J)
    # Labels
    logger.info("Labels are initialized by setting active probabilities "
                "to ones ...")
    #q_Z = np.ones((M, K, J), dtype=np.float64) / 2.
    q_Z = np.zeros((M, K, J), dtype=np.float64)
    q_Z[:, 1, :] = 1
    q_Z1 = copy.deepcopy(q_Z)
    Z_tilde = copy.deepcopy(q_Z)

    # H and G
    TT, m_h = getCanoHRF(Thrf, dt)
    H = np.array(m_h[:D]).astype(np.float64)
    H /= np.linalg.norm(H)
    Hb = create_physio_brf(phy_params, response_dt=dt, response_duration=Thrf)
    Hb /= np.linalg.norm(Hb)
    if prior=='balloon':
        H = Hb.copy()
    H1 = copy.deepcopy(H)
    Sigma_H = np.zeros((D, D), dtype=np.float64)

    # Initialize model parameters
    Beta = beta * np.ones((M), dtype=np.float64)
    P = vt.PolyMat(N, 4, TR)
    L = vt.polyFit(Y, TR, 4, P)
    WP = P.copy() #np.append(w[:, np.newaxis], P, axis=1)
    AL = L.copy() #np.append(alpha[np.newaxis, :], L, axis=0)
    y_tilde = Y - WP.dot(AL)

    # Parameters Gaussian mixtures
    mu_Ma = np.append(np.zeros((M, 1)), np.ones((M, 1)), axis=1).astype(np.float64)
    sigma_Ma = np.ones((M, K), dtype=np.float64) * 0.3

    # Params RLs
    m_A = np.zeros((J, M), dtype=np.float64)
    for j in xrange(0, J):
        m_A[j, :] = (np.random.normal(mu_Ma, np.sqrt(sigma_Ma)) * q_Z[:, :, j]).sum(axis=1).T
    m_A1 = m_A.copy()
    Sigma_A = np.ones((M, M, J)) * np.identity(M)[:, :, np.newaxis]

    G = np.zeros_like(H)
    m_C = np.zeros_like(m_A)
    Sigma_G = np.zeros_like(Sigma_H)
    Sigma_C = np.zeros_like(Sigma_A)
    mu_Mc = np.zeros_like(mu_Ma)
    sigma_Mc = np.ones_like(sigma_Ma)
    W = np.zeros_like(Gamma)


    # Precomputations
    WX = W.dot(XX).transpose(1, 0, 2)
    Gamma_X = np.tensordot(Gamma, XX, axes=(1, 1))
    X_Gamma_X = np.tensordot(XX.T, Gamma_X, axes=(1, 0))    # shape (D, M, M, D)
    Gamma_WX = np.tensordot(Gamma, WX, axes=(1, 1))
    XW_Gamma_WX = np.tensordot(WX.T, Gamma_WX, axes=(1, 0))    # shape (D, M, M, D)
    Gamma_WP = Gamma.dot(WP)
    WP_Gamma_WP = WP.T.dot(Gamma_WP)
    sigma_eps_m = np.maximum(sigma_eps, eps)
    cov_noise = sigma_eps_m[:, np.newaxis, np.newaxis]

    ###########################################################################
    #############################################             VBJDE

    t1 = time.time()
    ni = 0

    #while ((ni < NitMin + 1) or (((Crit_AH > Thresh) or (Crit_CG > Thresh)) \
    #        and (ni < NitMax))):
    #while ((ni < NitMin + 1) or (((Crit_AH > Thresh)) \
    #        and (ni < NitMax))):
    while ((ni < NitMin + 1) or (((Crit_FE > Thresh * np.ones_like(Crit_FE)).any()) \
            and (ni < NitMax))):

        logger.info("-------- Iteration n° " + str(ni + 1) + " --------")

        if PLOT and ni >= 0:  # Plotting HRF and PRF
            logger.info("Plotting HRF and PRF for current iteration")
            vt.plot_response_functions_it(ni, NitMin, M, H, G)


        # Managing types of prior
        priorH_cov_term = np.zeros_like(R_inv)
        matrix_covH = R_inv.copy()
        if prior=='balloon':
            logger.info("   prior balloon")
            #matrix_covH = np.eye(R_inv.shape[0], R_inv.shape[1])
            priorH_mean_term = np.dot(matrix_covH / sigmaH, Hb)
        else:
            logger.info("   NO prior")
            priorH_mean_term = np.zeros_like(H)
            priorG_mean_term = np.zeros_like(G)


        #####################
        # EXPECTATION
        #####################


        # HRF H
        if estimateH:
            logger.info("E H step ...")
            Ht, Sigma_H = vt.expectation_H_asl(Sigma_A, m_A, m_C, G, XX, W, Gamma,
                                            Gamma_X, X_Gamma_X, J, y_tilde,
                                            cov_noise, matrix_covH, sigmaH,
                                            priorH_mean_term, priorH_cov_term)

            if constraint:
                if not np.linalg.norm(Ht)==1:
                    logger.info("   constraint l2-norm = 1")
                    H = vt.constraint_norm1_b(Ht, Sigma_H)
                    #H = Ht / np.linalg.norm(Ht)
                else:
                    logger.info("   l2-norm already 1!!!!!")
                    H = Ht.copy()
                    #H[0] = 0
                    #H[-1] = 0
                Sigma_H = np.zeros_like(Sigma_H)
            else:
                H = Ht.copy()
                #H[0] = 0
                #H[-1] = 0
                h_norm = np.append(h_norm, np.linalg.norm(H))
                print 'h_norm = ', h_norm

            Crit_H = (np.linalg.norm(H - H1) / np.linalg.norm(H1)) ** 2
            cH += [Crit_H]
            H1[:] = H[:]


        # A
        if estimateA:
            logger.info("E A step ...")
            m_A, Sigma_A = vt.expectation_A_asl(H, G, m_C, W, XX, Gamma, Gamma_X, q_Z,
                                             mu_Ma, sigma_Ma, J, y_tilde,
                                             Sigma_H, sigma_eps_m)

            cA += [(np.linalg.norm(m_A - m_A1) / np.linalg.norm(m_A1)) ** 2]
            m_A1[:, :] = m_A[:, :]

        # Q labels
        if estimateZ:
            logger.info("E Q step ...")
            q_Z, Z_tilde = vt.expectation_Q_asl(Sigma_A, m_A, Sigma_C, m_C,
                                            sigma_Ma, mu_Ma, sigma_Mc, mu_Mc,
                                            Beta, Z_tilde, q_Z, neighboursIndexes, graph, M, J, K)

            cZ += [(np.linalg.norm(q_Z - q_Z1) / (np.linalg.norm(q_Z1) + eps)) ** 2]
            q_Z1 = q_Z


        if ni > 0:
            _, _, _, free_energyE = vt.Compute_FreeEnergy(y_tilde, m_A, Sigma_A, mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C, Sigma_C, mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps, XX, W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X, Gamma_WX)
            if free_energyE < free_energy:
                logger.info("free energy has decreased after E step from %f to %f", free_energy, free_energyE)


        # crit. AH and CG
        logger.info("crit. AH and CG")
        AH = m_A[:, :, np.newaxis] * H[np.newaxis, np.newaxis, :]

        Crit_AH = (np.linalg.norm(AH - AH1) / (np.linalg.norm(AH1) + eps)) ** 2
        cAH += [Crit_AH]
        AH1 = AH.copy()
        logger.info("Crit_AH = " + str(Crit_AH))


        #####################
        # MAXIMIZATION
        #####################

        if prior=='balloon':
            logger.info("   prior balloon")
            AuxH = H - Hb
            AuxG = G - Gb
        else:
            logger.info("   NO prior")
            AuxH = H.copy()
            AuxG = G.copy()

        # Variance HRF: sigmaH
        if estimateSigmaH:
            logger.info("M sigma_H step ...")
            sigmaH = vt.maximization_sigma_asl(D, Sigma_H, matrix_covH, AuxH, use_hyperprior, gamma_h)
            logger.info('sigmaH = ' + str(sigmaH))

        if ni > 0:
            _, _, _, free_energyVh = vt.Compute_FreeEnergy(y_tilde, m_A, Sigma_A, mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C, Sigma_C, mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps, XX, W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X, Gamma_WX)

            if free_energyVh < free_energyE:
                logger.info("free energy has decreased after v_h computation from %f to %f", free_energyE, free_energyVh)


        # (mu,sigma)
        if estimateMP:
            logger.info("M (mu,sigma) a and c step ...")
            mu_Ma, sigma_Ma = vt.maximization_mu_sigma_asl(q_Z, m_A, Sigma_A)

        if ni > 0:
            _, _, _, free_energyMP = vt.Compute_FreeEnergy(y_tilde, m_A, Sigma_A, mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C, Sigma_C, mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps, XX, W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X, Gamma_WX)
            if free_energyMP < free_energyVh:
                logger.info("free energy has decreased after GMM parameters computation from %f to %f", free_energyVh, free_energyMP)


        # Drift L, alpha
        if estimateLA:
            logger.info("M L, alpha step ...")
            AL = vt.maximization_LA_asl(Y, m_A, m_C, XX, WP, W, WP_Gamma_WP, H, G, Gamma)
            y_tilde = Y - WP.dot(AL)

        if ni > 0:
            _, _, _, free_energyLA = vt.Compute_FreeEnergy(y_tilde, m_A, Sigma_A, mu_Ma, sigma_Ma,
                                                 H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                                 m_C, Sigma_C, mu_Mc, sigma_Mc, G, Sigma_G,
                                                 AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                                 gamma, gamma_h, gamma_g, sigma_eps, XX, W,
                                                 J, D, M, N, K, use_hyperprior, Gamma_X, Gamma_WX)
            if free_energyLA < free_energyMP:
                logger.info("free energy has decreased after drifts computation from %f to %f", free_energyMP, free_energyLA)


        # Beta
        if estimateBeta:
            logger.info("M beta step ...")
            """Qtilde = np.concatenate((Z_tilde, np.zeros((M, K, 1), dtype=Z_tilde.dtype)), axis=2)
            Qtilde_sumneighbour = Qtilde[:, :, neighboursIndexes].sum(axis=3)
            Beta = vt.maximization_beta_m2(Beta.copy(), q_Z, Qtilde_sumneighbour,
                                             Qtilde, neighboursIndexes, maxNeighbours,
                                             gamma, MaxItGrad, gradientStep)
            logger.info(Beta)
            """
            logger.info("M beta step ...")
            Qtilde = np.concatenate((Z_tilde, np.zeros((M, K, 1), dtype=Z_tilde.dtype)), axis=2)
            Qtilde_sumneighbour = Qtilde[:, :, neighboursIndexes].sum(axis=3)
            for m in xrange(0, M):
                Beta[m] = vt.maximization_beta_m2_scipy_asl(Beta[m].copy(), q_Z[m, :, :], Qtilde_sumneighbour[m, :, :],
                                                   Qtilde[m, :, :], neighboursIndexes, maxNeighbours,
                                                   gamma, MaxItGrad, gradientStep)
            logger.info(Beta)
        if ni > 0:
            _, _, _, free_energyB = vt.Compute_FreeEnergy(y_tilde, m_A, Sigma_A, mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C, Sigma_C, mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps, XX, W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X, Gamma_WX)
            if free_energyB < free_energyLA:
                logger.info("free energy has decreased after Beta computation from %f to %f", \
                                free_energyLA, free_energyB)
        if 0 and ni < 5:
            plt.close('all')
            for m in xrange(0, M):
                range_b = np.arange(-10., 20., 0.1)
                beta_plotting = np.zeros_like(range_b)
                grad_plotting = np.zeros_like(range_b)
                for ib, b in enumerate(range_b):
                    beta_plotting[ib] = vt.fun(b, q_Z[m, :, :], Qtilde_sumneighbour[m, :, :],
                                                          neighboursIndexes, gamma)
                    grad_plotting[ib] = vt.grad_fun(b, q_Z[m, :, :], Qtilde_sumneighbour[m, :, :],
                                                     neighboursIndexes, gamma)
                #print beta_plotting
                plt.figure(1)
                plt.hold('on')
                plt.plot(range_b, beta_plotting)
                plt.figure(2)
                plt.hold('on')
                plt.plot(range_b, grad_plotting)
            plt.show()


        # Sigma noise
        if estimateNoise:
            logger.info("M sigma noise step ...")
            sigma_eps = vt.maximization_sigma_noise_asl(XX, m_A, Sigma_A, H, m_C, Sigma_C, \
                                                    G, Sigma_H, Sigma_G, W, y_tilde, Gamma, \
                                                    Gamma_X, Gamma_WX, N)

        if PLOT:
            for m in xrange(M):
                SUM_q_Z[m] += [q_Z[m, 1, :].sum()]
                mua1[m] += [mu_Ma[m, 1]]


        EPt, EPt_lh, Entropy, free_energy = vt.Compute_FreeEnergy(y_tilde, m_A, Sigma_A, mu_Ma, sigma_Ma,
                                             H, Sigma_H, AuxH, R, R_inv, sigmaH, sigmaG,
                                             m_C, Sigma_C, mu_Mc, sigma_Mc, G, Sigma_G,
                                             AuxG, q_Z, neighboursIndexes, Beta, Gamma,
                                             gamma, gamma_h, gamma_g, sigma_eps, XX, W,
                                             J, D, M, N, K, use_hyperprior, Gamma_X, Gamma_WX,
                                             plot=True, bold=True)
        if ni > 0:
            if free_energy < free_energyB:
                logger.info("free energy has decreased after Noise computation from %f to %f", free_energyB, free_energy)

        if ni > 0:
            if free_energy < FE[-1]:
                logger.info("WARNING! free energy has decreased in this iteration from %f to %f", FE[-1], free_energy)

        FE += [free_energy]
        EP += [EPt]
        EPlh += [EPt_lh]
        Ent += [Entropy]

        if ni > 5:
            #Crit_FE = np.abs((FE[-1] - FE[-2]) / FE[-2])
            FE0 = np.array(FE)
            Crit_FE = np.abs((FE0[-5:] - FE0[-6:-1]) / FE0[-6:-1])
            print Crit_FE
            print (Crit_FE > Thresh * np.ones_like(Crit_FE)).any()
        else:
            Crit_FE = 100

        ni += 1
        cTime += [time.time() - t1]

        logger.info("Computing reconstruction error")
        StimulusInducedSignal = vt.computeFit_asl(H, m_A, G, m_C, W, XX)
        rerror = np.append(rerror, \
                           np.mean(((Y - StimulusInducedSignal) ** 2).sum(axis=0)) \
                           / np.mean((Y ** 2).sum(axis=0)))

    CompTime = time.time() - t1


    # Normalize if not done already
    if not constraint: # or not normg:
        logger.info("l2-norm of H and G to 1 if not constraint")
        Hnorm = np.linalg.norm(H)
        H /= Hnorm
        Sigma_H /= Hnorm**2
        m_A *= Hnorm


    ## Compute contrast maps and variance
    if computeContrast and len(contrasts) > 0:
        logger.info("Computing contrasts ... ")
        CONTRAST_A, CONTRASTVAR_A, \
        CONTRAST_C, CONTRASTVAR_C = vt.compute_contrasts(condition_names,
                                                         contrasts, m_A, m_C,
                                                         Sigma_A, Sigma_C, M, J)
    else:
        CONTRAST_A, CONTRASTVAR_A, CONTRAST_C, CONTRASTVAR_C = 0, 0, 0, 0


    ###########################################################################
    ##########################################    PLOTS and SNR computation

    logger.info("Nb iterations to reach criterion: %d",  ni)
    logger.info("Computational time = %s min %s s",
                str(np.int(CompTime // 60)), str(np.int(CompTime % 60)))
    logger.info("Iteration time = %s min %s s",
                str(np.int((CompTime // ni) // 60)), str(np.int((CompTime / ni) % 60)))

    logger.info("perfusion baseline mean = %f", np.mean(AL[0, :]))
    logger.info("perfusion baseline var = %f", np.var(AL[0, :]))
    logger.info("drifts mean = %f", np.mean(AL[1:, :]))
    logger.info("drifts var = %f", np.var(AL[1:, :]))
    logger.info("noise mean = %f", np.mean(sigma_eps))
    logger.info("noise var = %f", np.var(sigma_eps))

    SNR10 = 20 * (np.log10(np.linalg.norm(Y) / \
                np.linalg.norm(Y - StimulusInducedSignal - WP.dot(AL))))
    logger.info("SNR = %d",  SNR10)

    return ni, m_A, H, m_C, G, Z_tilde, sigma_eps, \
           mu_Ma, sigma_Ma, mu_Mc, sigma_Mc, Beta, AL[:, :], np.dot(P, AL), \
           np.zeros_like(AL[0, :]), Sigma_A, Sigma_C, Sigma_H, Sigma_G, rerror, \
           CONTRAST_A, CONTRASTVAR_A, CONTRAST_C, CONTRASTVAR_C, \
           cA[:], cH[2:], cC[2:], cG[2:], cZ[2:], cAH[2:], cCG[2:], \
           cTime, FE, EP, EPlh, Ent




# -*- coding: utf-8 -*-

import os
import os.path as op
import unittest
import shutil
import logging

from copy import deepcopy

import numpy as np

import pyhrf
import pyhrf.boldsynth.scenarios as sim

from pyhrf.jde.jde_multi_sujets import BOLDGibbs_Multi_SubjSampler as BMSS
from pyhrf.jde.jde_multi_sujets import simulate_single_subject
from pyhrf.jde import jde_multi_sujets as jms
from pyhrf.ui.jde import JDEMCMCAnalyser
from pyhrf.ui.treatment import FMRITreatment
from pyhrf import Condition
from pyhrf.core import FmriData, FmriGroupData


logger = logging.getLogger(__name__)


def simulate_subjects(output_dir, snr_scenario='high_snr',
                      spatial_size='tiny', hrf_group=None, nb_subjects=15,
                      vhrf=0.1, vhrf_group=0.1):
    '''
    Simulate daata for multiple subjects (5 subjects by default)
    '''
    drift_coeff_var = 1.
    drift_amplitude = 10.

    lmap1, lmap2, lmap3 = 'random_small', 'random_small', 'random_small'

    if snr_scenario == 'low_snr':  # low snr
        vars_noise = np.zeros(nb_subjects) + 1.5
        conditions = [
            Condition(name='audio', m_act=3., v_act=.3, v_inact=.3,
                      label_map=lmap1),
            Condition(name='video', m_act=2.5, v_act=.3, v_inact=.3,
                      label_map=lmap2),
            Condition(name='damier', m_act=2, v_act=.3, v_inact=.3,
                      label_map=lmap3),
        ]
    else:  # high snr

        vars_noise = np.zeros(nb_subjects) + .2
        conditions = [
            Condition(name='audio', m_act=13., v_act=.2, v_inact=.1,
                      label_map=lmap1),
            # Condition(name='video', m_act=11.5, v_act=.2, v_inact=.1,
            # label_map=lmap2),
            # Condition(name='damier', m_act=10, v_act=.2, v_inact=.1,
            # label_map=lmap3),
        ]
    vars_hrfs = np.zeros(nb_subjects) + vhrf

    # Common variable across subjects:
    labels_vol = sim.create_labels_vol(conditions)
    labels = sim.flatten_labels_vol(labels_vol)

    # use smooth multivariate gaussian prior:
    if hrf_group is None:  # simulate according to gaussian prior
        var_hrf_group = 0.1
        hrf_group = sim.create_gsmooth_hrf(dt=0.6, hrf_var=var_hrf_group,
                                           normalize_hrf=False)
        n = (hrf_group ** 2).sum() ** .5
        hrf_group /= n
        var_hrf_group /= n ** 2

    simu_subjects = []
    simus = []

    for isubj in xrange(nb_subjects):
        if output_dir is not None:
            out_dir = op.join(output_dir, 'subject_%d' % isubj)
            if not op.exists(out_dir):
                os.makedirs(out_dir)
        else:
            out_dir = None
        s = simulate_single_subject(out_dir, conditions, vars_hrfs[isubj],
                                    labels, labels_vol, vars_noise[isubj],
                                    drift_coeff_var,
                                    drift_amplitude, hrf_group, dt=0.6, dsf=4,
                                    var_hrf_group=vhrf_group)
        if 0:
            print 'simu subj %d:' % isubj
            print 'vhs:', s['var_subject_hrf']
            print 'hg:', s['hrf_group']
            print 'vhg:', s['var_hrf_group']

        simus.append(s)
        simu_subjects.append(FmriData.from_simulation_dict(s))
    simu_subjects = FmriGroupData(simu_subjects)

    return simu_subjects


class MultiSubjTest(unittest.TestCase):

    def setUp(self):
        np.random.seed(8652761)

        self.simu_dir = pyhrf.get_tmp_path()

        # Parameters to setup a Sampler where all samplers are OFF and
        # set to their true values.
        # This is used by the function _test_specific_samplers,
        # which will turn on specific samplers to test.

        self.sampler_params_for_single_test = {
            'nb_iterations': 40,
            'smpl_hist_pace': -1,
            'obs_hist_pace': -1,
            # HRF by subject
            'hrf_subj': jms.HRF_Sampler(do_sampling=False,
                                        normalise=1.,
                                        use_true_value=True,
                                        zero_contraint=False,
                                        prior_type='singleHRF'),

            # HRF variance
            'hrf_var_subj': jms.HRFVarianceSubjectSampler(do_sampling=False,
                                                          use_true_value=True),
            # HRF group
            'hrf_group': jms.HRF_Group_Sampler(do_sampling=False,
                                               normalise=1.,
                                               use_true_value=True,
                                               zero_contraint=False,
                                               prior_type='singleHRF'),
            # HRF variance
            'hrf_var_group': jms.RHGroupSampler(do_sampling=False,
                                                use_true_value=True),

            # neural response levels (stimulus-induced effects) by subject
            'response_levels': jms.NRLs_Sampler(do_sampling=False,
                                                use_true_value=True),
            'labels': jms.LabelSampler(do_sampling=False,
                                       use_true_value=True),
            # drift
            'drift': jms.Drift_MultiSubj_Sampler(do_sampling=False,
                                                 use_true_value=True),
            # drift variance
            'drift_var': jms.ETASampler_MultiSubj(do_sampling=False,
                                                  use_true_value=True),
            # noise variance
            'noise_var':
            jms.NoiseVariance_Drift_MultiSubj_Sampler(do_sampling=False,
                                                      use_true_value=False),
            # weights o fthe mixture
            # parameters of the mixture
            'mixt_params': jms.MixtureParamsSampler(do_sampling=False,
                                                    use_true_value=False),
            #'alpha_subj' : Alpha_hgroup_Sampler(dict_alpha_single),
            #'alpha_var_subj' : AlphaVar_Sampler(dict_alpha_var_single),
            'check_final_value': 'none',  # print or raise
        }

    def tearDown(self):
        shutil.rmtree(self.simu_dir)

    def _test_specific_samplers(self, sampler_names, simu,
                                nb_its=None, use_true_val=None,
                                save_history=False, check_fv=None,
                                normalize_hrf=1.,
                                hrf_prior_type='singleHRF',
                                normalize_hrf_group=1.,
                                hrf_group_prior_type='singleHRF',
                                reg_hgroup=True):
        """
        Test specific samplers.
        """
        if use_true_val is None:
            use_true_val = dict((n, False) for n in sampler_names)

        logger.info('_test_specific_samplers %s ...', str(sampler_names))

        params = deepcopy(self.sampler_params_for_single_test)

        # Loop over given samplers to enable them
        for var_name in sampler_names:
            var_class = params[var_name].__class__
            use_tval = use_true_val[var_name]

            # special case for HRF -> normalization and prior type
            if var_class == jms.HRF_Sampler:
                params[var_name] = jms.HRF_Sampler(do_sampling=True,
                                                   use_true_value=use_tval,
                                                   normalise=normalize_hrf,
                                                   prior_type=hrf_prior_type,
                                                   zero_contraint=False)
            elif var_class == jms.HRF_Group_Sampler:
                ptype = hrf_group_prior_type
                nhg = normalize_hrf_group
                shg = jms.HRF_Group_Sampler(do_sampling=True,
                                            use_true_value=use_tval,
                                            normalise=nhg,
                                            prior_type=ptype,
                                            zero_contraint=False,
                                            regularise=reg_hgroup)
                params[var_name] = shg
            else:
                params[var_name] = var_class(do_sampling=True,
                                             use_true_value=use_tval)

        if nb_its is not None:
            params['nb_iterations'] = nb_its

        if save_history:
            params['smpl_hist_pace'] = 1
            params['obs_hist_pace'] = 1

        if check_fv is not None:
            params['check_final_value'] = check_fv

        sampler = BMSS(**params)

        output_dir = self.simu_dir

        analyser = JDEMCMCAnalyser(sampler=sampler, osfMax=4, dtMin=.4,
                                   dt=.6, driftParam=4, driftType='polynomial',
                                   outputPrefix='jde_mcmc_',
                                   randomSeed=5421087, pass_error=False)

        treatment = FMRITreatment(fmri_data=simu, analyser=analyser,
                                  output_dir=output_dir)

        outputs = treatment.run()
        return outputs

    def test_quick(self):
        """ Test running of JDE multi subject (do not test result accuracy) """
        simu = simulate_subjects(self.simu_dir, nb_subjects=2)
        self._test_specific_samplers(['hrf_subj'], simu, nb_its=2,
                                     check_fv=None)
